IMPLEMENTATION MODULE ScrDir;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **************************************************************************
    Compiler:   TopSpeed / Stony Brook Modula-2
    --------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Initial version                         November 94
    A.Iakovlev  Ver  1.1  Acceleration                            February 95
    A.Iakovlev  Ver  1.2  Use variable ScrExt.bgr                 February 95
    **************************************************************************
    TopSpeed compiler dependencies:
    ------ structured constants
    TopSpeed library dependencies:
    ------ import from SYSTEM: BYTE
    Stony Brook compiler dependencies:
    ------ structured constants
    Stony Brook library dependencies:
    ------ import from SYSTEM: BYTE
  *)

  FROM Cfg IMPORT CMPTS, CMPSB;
  IMPORT SYSTEM, ScrExt, Scr;

  (* ====================================== Global types *)

  TYPE RGB = ARRAY Scr.PureColor OF CARDINAL;

  (* ====================================== 16-color functions *)

  (*%T CMPTS*)

  CONST
    DirectShades4 =   RGB( 2H,  4H,  2H); (* --- number of shades *)
    BitMask4      =   RGB( 8H,  6H,  1H); (* --- bit mask for shade *)
    Mult4         =   RGB( 8H,  2H,  1H); (* --- multiplier for left shift *)
    Inc4          =   RGB(80H, 40H, 80H); (* --- MaxShades DIV DirectShades *)
    CombMult4     =   RGB(10H, 20H, 80H); (* --- Inc DIV Mult *)

  (*%E*)

  (*%T CMPSB*)

  CONST
    DirectShades4 :RGB = [ 2H,  4H,  2H]; (* --- number of shades *)
    BitMask4      :RGB = [ 8H,  6H,  1H]; (* --- bit mask for shade *)
    Mult4         :RGB = [ 8H,  2H,  1H]; (* --- multiplier for left shift *)
    Inc4          :RGB = [80H, 40H, 80H]; (* --- MaxShades DIV DirectShades *)
    CombMult4     :RGB = [10H, 20H, 80H]; (* --- Inc DIV Mult *)

  (*%E*)

  PROCEDURE DirectShadesNmb4(pc: Scr.PureColor): CARDINAL;
    BEGIN
      RETURN DirectShades4[pc];
    END DirectShadesNmb4;

  PROCEDURE SetDirectShade4(VAR c: Scr.Pixel; pc: Scr.PureColor; sh: CARDINAL);
    VAR
      temp: BITSET;
    BEGIN
      temp :=
        (BITSET(VAL(CARDINAL, c)) - BITSET(BitMask4[pc])) +
        (BITSET(sh DIV CombMult4[pc]) * BITSET(BitMask4[pc]));
      c := VAL(Scr.Pixel, CARDINAL(temp));
    END SetDirectShade4;

  PROCEDURE DirectShade4(c: Scr.Pixel; pc: Scr.PureColor): CARDINAL;
    VAR
      temp: BITSET;
    BEGIN
      temp := BITSET(VAL(CARDINAL, c)) * BITSET(BitMask4[pc]);
      RETURN CARDINAL(temp) * CombMult4[pc];
    END DirectShade4;

  (* ============================================ 256-color functions *)

  (*%T CMPTS*)

  CONST
    DirectShades8 =   RGB(  8H,  8H,  4H); (* --- number of shades *)
    BitMask8      =   RGB(0E0H, 1CH,  3H); (* --- bit mask for shade *)
    Mult8         =   RGB( 20H,  4H,  1H); (* --- multiplier for left shift *)
    Inc8          =   RGB( 20H, 20H, 40H); (* --- MaxShades DIV DirectShades *)
    CombMult8     =   RGB(  1H,  8H, 40H); (* --- Inc DIV Mult *)

  (*%E*)

  (*%T CMPSB*)

  CONST
    DirectShades8 :RGB = [  8H,  8H,  4H]; (* --- number of shades *)
    BitMask8      :RGB = [0E0H, 1CH,  3H]; (* --- bit mask for shade *)
    Mult8         :RGB = [ 20H,  4H,  1H]; (* --- multiplier for left shift *)
    Inc8          :RGB = [ 20H, 20H, 40H]; (* --- MaxShades DIV DirectShades *)
    CombMult8     :RGB = [  1H,  8H, 40H]; (* --- Inc DIV Mult *)

  (*%E*)

  PROCEDURE DirectShadesNmb8(pc: Scr.PureColor): CARDINAL;
    BEGIN
      RETURN DirectShades8[pc];
    END DirectShadesNmb8;

  PROCEDURE SetDirectShade8(VAR c: Scr.Pixel; pc: Scr.PureColor; sh: CARDINAL);
    VAR
      temp: BITSET;
    BEGIN
      temp :=
        (BITSET(VAL(CARDINAL, c)) - BITSET(BitMask8[pc])) +
        (BITSET(sh DIV CombMult8[pc]) * BITSET(BitMask8[pc]));
      c := VAL(Scr.Pixel, CARDINAL(temp));
    END SetDirectShade8;

  PROCEDURE DirectShade8(c: Scr.Pixel; pc: Scr.PureColor): CARDINAL;
    VAR
      temp: BITSET;
    BEGIN
      temp := BITSET(VAL(CARDINAL, c)) * BITSET(BitMask8[pc]);
      RETURN CARDINAL(temp) * CombMult8[pc];
    END DirectShade8;

  (* ============================================== 32K-color functions *)

  CONST
    DirectShades15 = 32;
    Inc15 = 8;

  (*%T CMPTS*)

  CONST
    BitMask15      =   RGB(7C00H, 3E0H, 1FH); (* --- bit mask for shade *)
    Mult15         =   RGB( 400H,  20H,  1H); (* --- multiplier for left shift *)

  (*%E*)

  (*%T CMPSB*)

  CONST
    BitMask15      :RGB = [7C00H, 3E0H, 1FH]; (* --- bit mask for shade *)
    Mult15         :RGB = [ 400H,  20H,  1H]; (* --- multiplier for left shift *)

  (*%E*)

  PROCEDURE DirectShadesNmb15(pc: Scr.PureColor): CARDINAL;
    BEGIN
      RETURN DirectShades15;
    END DirectShadesNmb15;

  PROCEDURE SetDirectShade15(VAR c: Scr.Pixel; pc: Scr.PureColor; sh: CARDINAL);
    VAR
      temp: BITSET;
    BEGIN
      temp :=
        (BITSET(VAL(CARDINAL, c)) - BITSET(BitMask15[pc])) +
        (BITSET((sh DIV Inc15) * Mult15[pc]) * BITSET(BitMask15[pc]));
      c := VAL(Scr.Pixel, CARDINAL(temp));
    END SetDirectShade15;

  PROCEDURE DirectShade15(c: Scr.Pixel; pc: Scr.PureColor): CARDINAL;
    VAR
      temp: BITSET;
    BEGIN
      temp := BITSET(VAL(CARDINAL, c)) * BITSET(BitMask15[pc]);
      RETURN (CARDINAL(temp) DIV Mult15[pc]) * Inc15;
    END DirectShade15;

  (* =============================================== 64K-color functions *)

  (*%T CMPTS*)

  CONST
    DirectShades16 =   RGB(   20H,  40H, 20H); (* --- number of shades *)
    BitMask16      =   RGB(0F800H, 7E0H, 1FH); (* --- bit mask for shade *)
    Mult16         =   RGB(  800H,  20H,  1H); (* --- multiplier for left shift *)
    Inc16          =   RGB(    8H,   4H,  8H); (* --- MaxShades DIV DirectShades *)

  (*%E*)

  (*%T CMPSB*)

  CONST
    DirectShades16 :RGB = [   20H,  40H, 20H]; (* --- number of shades *)
    BitMask16      :RGB = [0F800H, 7E0H, 1FH]; (* --- bit mask for shade *)
    Mult16         :RGB = [  800H,  20H,  1H]; (* --- multiplier for left shift *)
    Inc16          :RGB = [    8H,   4H,  8H]; (* --- MaxShades DIV DirectShades *)

  (*%E*)

  PROCEDURE DirectShadesNmb16(pc: Scr.PureColor): CARDINAL;
    BEGIN
      RETURN DirectShades16[pc];
    END DirectShadesNmb16;

  PROCEDURE SetDirectShade16(VAR c: Scr.Pixel; pc: Scr.PureColor; sh: CARDINAL);
    VAR
      temp: BITSET;
    BEGIN
      temp :=
        (BITSET(VAL(CARDINAL, c)) - BITSET(BitMask16[pc])) +
        (BITSET((sh DIV Inc16[pc]) * Mult16[pc]) * BITSET(BitMask16[pc]));
      c := VAL(Scr.Pixel, CARDINAL(temp));
    END SetDirectShade16;

  PROCEDURE DirectShade16(c: Scr.Pixel; pc: Scr.PureColor): CARDINAL;
    VAR
      temp: BITSET;
    BEGIN
      temp := BITSET(VAL(CARDINAL, c)) * BITSET(BitMask16[pc]);
      RETURN (CARDINAL(temp) DIV Mult16[pc]) * Inc16[pc];
    END DirectShade16;

  (* ============================================= 16M-color functions *)

  CONST
    DirectShades24 = 256;

  TYPE
    Pixel =
      RECORD
        CASE : CARDINAL OF
        | 0:
          pixel: Scr.Pixel;
        | 1:
          acolor: ARRAY Scr.PureColor OF SYSTEM.BYTE;
        END;
      END;

  PROCEDURE DirectShadesNmb24(pc: Scr.PureColor): CARDINAL;
    BEGIN
      RETURN DirectShades24;
    END DirectShadesNmb24;

  PROCEDURE SetDirectShade24(VAR c: Scr.Pixel; pc: Scr.PureColor; sh: CARDINAL);
    VAR
      ptr: POINTER TO Pixel;
    BEGIN
      IF ScrExt.bgr THEN 
        pc := VAL(Scr.PureColor, ORD(MAX(Scr.PureColor)) - ORD(pc));
      END;
      ptr := SYSTEM.ADR(c);
      ptr^.acolor[pc] := SYSTEM.BYTE(sh);
    END SetDirectShade24;

  PROCEDURE DirectShade24(c: Scr.Pixel; pc: Scr.PureColor): CARDINAL;
    VAR
      ptr: POINTER TO Pixel;
    BEGIN
      IF ScrExt.bgr THEN 
        pc := VAL(Scr.PureColor, ORD(MAX(Scr.PureColor)) - ORD(pc));
      END;
      ptr := SYSTEM.ADR(c);
      RETURN ORD(ptr^.acolor[pc]);
    END DirectShade24;

  END ScrDir.
