IMPLEMENTATION MODULE ScrOut;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **************************************************************************
    Compilers:  TopSpeed / Stony Brook Modula-2
    --------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0                                          March 89
    A.Iakovlev  Ver  2.0  Renaming                                January 90
    A.Iakovlev  Ver  3.0  Attributes by value                     February 90
    A.Iakovlev  Ver  4.0  Changes in attributes                   February 90
    A.Iakovlev  Ver  5.0  New module Scr (no mode in prims)       February 90
    A.Iakovlev  Ver  6.0  Renaming                                February 90
    A.Iakovlev  Ver  7.0  No moving positions                     March 90
    A.Iakovlev  Ver  8.0  New Scr & changes in attributes         June 90
    A.Iakovlev  Ver  9.0  Fonts management                        June 90
    A.Iakovlev  Ver 10.0  Return status in New/Dispose            June 90
    A.Iakovlev  Ver 10.1  Maximum use of ROM-based fonts          July 90
    A.Iakovlev  Ver 10.2  Conditional compilation for MS-DOS/OS2  July 90
    A.Iakovlev  Ver 10.3  Use switch module OpSystem              July 90
    A.Iakovlev  Ver 11.0  More comments and simplifications       July 90
    A.Iakovlev  Ver 12.0  Renaming and hot spot in attributes     October 90
    A.Iakovlev  Ver 13.0  Procedure PattRect and pattern font     October 90
    A.Iakovlev  Ver 14.0  New checks => Status, Dispose changed   October 90
    A.Iakovlev  Ver 14.1  Procedure Prev => module SrvObj         January 91
    A.Iakovlev  Ver 15.0  Procs for partial font description      January 91
    A.Iakovlev  Ver 15.1  Correction in New                       January 91
    A.Iakovlev  Ver 16.0  Renaming pattern -> hatch               February 91
    A.Iakovlev  Ver 16.1  Use module ScrBase                      May 91
    A.Iakovlev  Ver 17.0  Repositioning of standard fonts         October 91
    A.Iakovlev  Ver 18.0  Use CHARSET's in fonts, no proc New     November 91
    A.Iakovlev  Ver 18.1  Correction in initialization procedure  January 92
    A.Iakovlev  Ver 18.2  Use %T DOS instead of %F OS2            January 92
    A.Iakovlev  Ver 18.3  No DOS or OS/2 specifics                January 92
    A.Iakovlev  Ver 18.4  New Scr.FontInfo                        August 92
    A.Iakovlev  Ver 19.0  Use module Res                          August 92
    A.Iakovlev  Ver 20.0  Pragma data(var_enum_size)              August 92
    A.Iakovlev  Ver 21.0  Font table & NIL font processing        May 93
    A.Iakovlev  Ver 22.0  Renaming                                June 93
    A.Iakovlev  Ver 23.0  Import from SYSTEM                      January 94
    A.Iakovlev  Ver 23.1  Use module SrvLow                       January 94
    A.Iakovlev  Ver 24.0  Own type CharSet                        January 94
    A.Iakovlev  Ver 24.1  Use SrvSys.Add instead of Lib.AddAddr   January 94
    A.Iakovlev  Ver 24.2  Use module Strings                      January 94
    A.Iakovlev  Ver 24.3  No use SHORTCARD type                   February 94
    A.Iakovlev  Ver 25.0  Procedure CorrectAttrs                  February 94
    A.Iakovlev  Ver 25.1  Stony Brook support                     February 94
    A.Iakovlev  Ver 26.0  Use Scr.Pixel                           September 94
    A.Iakovlev  Ver 26.1  Use module Mem for heap control         October 94
    A.Iakovlev  Ver 27.0  Simplifications in Attrs                March 95
    A.Iakovlev  Ver 28.0  Attributes settings procedures          March 95
    A.Iakovlev  Ver 29.0  United output attributes/primitives     March 95
    **************************************************************************
    TopSpeed compiler dependencies:
    ------ options (*$...*) and pragmas (*#...*)
    ------ structured constants
    TopSpeed library dependencies:
    ------ import from SYSTEM: BYTE
    ------ import from Strings: Length
    Stony Brook compiler dependencies:
    ------ structured constants
    Stony Brook library dependencies:
    ------ import from SYSTEM: BYTE
    ------ import from Strings: Length
  *)

  FROM Cfg IMPORT CMPTS, CMPSB;
  IMPORT
    SYSTEM, Strings,
    Geo, Res, SrvSys, SrvLow, SrvObj, DrvMem, ScrBas, Scr;

  (* ====================================== Font type descriptor *)

  TYPE
    FONT = POINTER TO FontDescriptor;
    FontHeader =
      RECORD
        charWdt: CARDINAL;             (* --- width of character in pixels *)
        charHgt: CARDINAL;             (* --- height of character in pixels *)
        byteCharWdt: CARDINAL;         (* --- width of character in bytes *)
        byteCharLen: CARDINAL;         (* --- length of character in bytes *)
        fontLen: CARDINAL;             (* --- length of font in bytes *)
        fontAddr: SYSTEM.ADDRESS;      (* --- address of font memory *)
      END;
    Entry = SYSTEM.ADDRESS;
    FontDescriptor =
      RECORD
        next: FONT;                    (* --- next font in font chain *)
        header: FontHeader;            (* --- font header *)
        stipple: ARRAY CHAR OF Entry;  (* --- character address's array *)
        standard: BOOLEAN;             (* --- TRUE, iff standard font *)
      END;

  CONST
    NullStipple = Entry(0);

  (* ============================================== Font table variable *)

  CONST
    FontTableLen = 16;

  VAR
    fontTable: ARRAY [0 .. FontTableLen - 1] OF FONT;

  (* =========================================== Attributes setting *)

  PROCEDURE SetMarkAttrs(VAR a: Attrs;
            func: Scr.Function; color: Scr.Pixel;
            style: MarkerStyle; size: CARDINAL);
    BEGIN
      a.function := func;
      a.color := color;
      a.markerStyle := style;
      a.markerSize := size;
      a.fillStyle := fillSolid;
    END SetMarkAttrs;

  PROCEDURE SetLineAttrs(VAR a: Attrs;
            func: Scr.Function; color: Scr.Pixel);
    BEGIN
      a.function := func;
      a.color := color;
      a.fillStyle := fillSolid;
    END SetLineAttrs;

  PROCEDURE SetFillAttrs(VAR a: Attrs;
            func: Scr.Function; color, auxColor: Scr.Pixel;
            style: FillStyle; font: FONT;
            origin: Geo.Point; stippleNo: CARDINAL);
    BEGIN
      SetTextAttrs(a, func, color, auxColor, style, font);
      a.fillStippleNo := stippleNo;
      a.fillOrigin := origin;
    END SetFillAttrs;

  PROCEDURE SetTextAttrs(VAR a: Attrs;
            func: Scr.Function; color, auxColor: Scr.Pixel;
            style: FillStyle; font: FONT);
    BEGIN
      a.function := func;
      a.color := color;
      a.auxColor := auxColor;
      a.fillStyle := style;
      IF font # NIL THEN a.font := font ELSE a.font := fontTable[0] END;
    END SetTextAttrs;

  (* =================================================== Drawing utilities *)

  TYPE
    Data =
      RECORD
        pixmap: Scr.PIXMAP;
        attrs: Attrs;
      END;
    DataPtr = POINTER TO Data;

  VAR
    nullPt: Geo.Point;  (* --- fill stipple origin is always null *)

  PROCEDURE Draw(r: Geo.Rectangle; d: DataPtr);
  (* --- draw filled/stippled rectangle; should be called by tracers *)
    TYPE
      FillStyleSet = SET OF FillStyle;
    VAR
      stippleAddr: SYSTEM.ADDRESS;
      font: FONT;
    BEGIN
      WITH d^ DO
        (* ------------------- solid fill/opaque background fill *)
        IF attrs.fillStyle = fillSolid THEN
          Scr.SolidRect(pixmap, r, attrs.color, attrs.function);
        ELSIF attrs.fillStyle = fillOpaqueStippled THEN
          Scr.SolidRect(pixmap, r, attrs.auxColor, attrs.function);
        END;
        (* ------------------------------------------- stipple fill iff need *)
        IF attrs.fillStyle IN FillStyleSet{fillStippled, fillOpaqueStippled} THEN
	  font := attrs.font;
          WITH font^ DO
            stippleAddr := stipple[CHR(attrs.fillStippleNo)];
            IF stippleAddr # NullStipple THEN
              Scr.StippledRect
              (
                pixmap, r,
                stippleAddr, header.byteCharWdt, header.charHgt,
                attrs.fillOrigin, attrs.color, attrs.function
              );
            END;
          END;
        END;
      END;
    END Draw;

  (* =================================================== Mark primitive *)

  PROCEDURE Marker(pm: Scr.PIXMAP; a: Attrs; p: Geo.Point);
    VAR
      r: Geo.Rectangle;
      halfSize, delta: CARDINAL;
    BEGIN
      (* --------------------- set halfsize parameter *)
      IF (a.markerStyle # dot) AND (a.markerSize > 0) THEN
        halfSize := (a.markerSize - 1) DIV 2;
      END;
      (* ------------------------------------------- drawing *)
      CASE a.markerStyle OF
      | dot:
        (* ------------------------------------------- dot mark *)
        Scr.Pt(pm, p, a.color, a.function);
      | plus:
        (* ------------- plus mark: horizontal and vertical lines *)
        IF halfSize > 0 THEN
          r.ixMin := p.ix - VAL(INTEGER, halfSize);
          r.ixMax := p.ix + VAL(INTEGER, halfSize);
          r.iyMin := p.iy;
          r.iyMax := r.iyMin;
          Scr.SolidRect(pm, r, a.color, a.function);
          r.ixMin := p.ix;
          r.ixMax := r.ixMin;
          r.iyMin := p.iy - VAL(INTEGER, halfSize);
          r.iyMax := p.iy + VAL(INTEGER, halfSize);
          Scr.SolidRect(pm, r, a.color, a.function);
        END;
        (* ------------------------------- central point *)
        Scr.Pt(pm, p, a.color, a.function);
      | asterisk:
        (* ---------------------------------- asterisk mark *)
        IF halfSize > 0 THEN
          (* --------------------------------- horizontal line *)
          r.ixMin := p.ix - VAL(INTEGER, halfSize);
          r.ixMax := p.ix + VAL(INTEGER, halfSize);
          r.iyMin := p.iy;
          r.iyMax := r.iyMin;
          Scr.SolidRect(pm, r, a.color, a.function);
          (* ------------------------------------ diagonal lines *)
          FOR delta := 2 TO halfSize BY 2 DO
            DEC(p.ix);
            DEC(p.iy);
            Scr.Pt(pm, p, a.color, a.function);
            INC(p.ix, delta);
            Scr.Pt(pm, p, a.color, a.function);
            INC(p.iy, delta);
            Scr.Pt(pm, p, a.color, a.function);
            DEC(p.ix, delta);
            Scr.Pt(pm, p, a.color, a.function);
            DEC(p.iy,  delta);
          END;
        ELSE
          (* --------- asterisk is degenerated into the point *)
          Scr.Pt(pm, p, a.color, a.function);
        END;
      | circle:
        (* -------------------------------------------- circle mark *)
        Ellipse(pm, a, p, halfSize, halfSize);
      | cross:
        (* ---------------------------------- diagonal cross mark *)
        IF halfSize > 0 THEN
          FOR delta := 2 TO halfSize * 2 BY 2 DO
            DEC(p.ix);
            DEC(p.iy);
            Scr.Pt(pm, p, a.color, a.function);
            INC(p.ix, delta);
            Scr.Pt(pm, p, a.color, a.function);
            INC(p.iy, delta);
            Scr.Pt(pm, p, a.color, a.function);
            DEC(p.ix, delta);
            Scr.Pt(pm, p, a.color, a.function);
            DEC(p.iy,  delta);
          END;
        ELSE
          Scr.Pt(pm, p, a.color, a.function);
        END;
      | square:
        (* ---------------------------------- square mark *)
        r.ixMin := p.ix - VAL(INTEGER, halfSize);
        r.ixMax := p.ix + VAL(INTEGER, halfSize);
        r.iyMin := p.iy - VAL(INTEGER, halfSize);
        r.iyMax := p.iy + VAL(INTEGER, halfSize);
        Rectangle(pm, a, r);
      END;
    END Marker;

  (* =========================================== Line/Polyline primitives *)

  PROCEDURE Line(pm: Scr.PIXMAP; a: Attrs; p1, p2: Geo.Point);
    VAR
      buffer: Data;
    BEGIN
      (* ---------------- set the buffer *)
      buffer.attrs := a;
      buffer.pixmap := pm;
      (* ---------------------------------------- drawing *)
      Geo.TraceLine
      (
	p1, p2,
	Geo.Action(Draw), SYSTEM.ADR(buffer)
      );
    END Line;

  PROCEDURE Polyline(pm: Scr.PIXMAP; a: Attrs;
            ps: ARRAY OF Geo.Point; first, last: CARDINAL; close: BOOLEAN): Res.Status;
    VAR
      buffer: Data;
      j: CARDINAL;
    BEGIN
      (* ------------------- set the buffer *)
      buffer.attrs := a;
      buffer.pixmap := pm;
      (* ------------------------------------- drawing *)
      IF first < last THEN
        (* -------------------------------------------- polyline *)
        FOR j := first + 1 TO last DO
          Geo.TraceLine
          (
            ps[j - 1], ps[j],
            Geo.Action(Draw), SYSTEM.ADR(buffer)
          );
        END;
        (* -------------------------------------------- closing iff need *)
        IF close THEN
          Geo.TraceLine
          (
            ps[last], ps[first],
            Geo.Action(Draw), SYSTEM.ADR(buffer)
          );
        END;
      ELSIF first = last THEN
        (* ------------------------- polyline is degenerated into the point *)
        Geo.TraceLine
        (
          ps[first], ps[last],
          Geo.Action(Draw), SYSTEM.ADR(buffer)
        );
      END;
      RETURN Res.done;
    END Polyline;

  PROCEDURE PolylineMaxSize(): CARDINAL;
    BEGIN
      RETURN 0;
    END PolylineMaxSize;

  (* =============================================== Rectangle primitive *)

  PROCEDURE Rectangle(pm: Scr.PIXMAP; a: Attrs; r: Geo.Rectangle);
    VAR
      buffer: Data;
    BEGIN
      (* ------------------- set the buffer *)
      buffer.attrs := a;
      buffer.pixmap := pm;
      (* --------------------------------------- drawing *)
      Geo.TraceRect
      (
        r,
        Geo.Action(Draw), SYSTEM.ADR(buffer)
      );
    END Rectangle;

  (* =========================================== Circle/Ellipse primitives *)

  PROCEDURE Ellipse(pm: Scr.PIXMAP; a: Attrs; p: Geo.Point; a0, b0: CARDINAL);
    VAR
      buffer: Data;
      r: Geo.Rectangle;
    BEGIN
      (* ------------------- set the buffer *)
      buffer.attrs := a;
      buffer.pixmap := pm;
      (* ------------------------------------- drawing *)
      IF a0 > 0 THEN
        (* ------------------- ellipse has non-zero x-halfaxis *)
        IF b0 > 0 THEN
          (* --------- ellipse has internity: check half-axes equality *)
          IF a0 = b0 THEN
            (* -------------------------------------------- it is circle *)
            Geo.TraceCircle
            (
              p, a0,
              Geo.Action(Draw), SYSTEM.ADR(buffer)
            );
          ELSE
            (* --------------------------------------------- it is ellipse *)
            Geo.TraceEllipse
            (
              p, a0, b0,
              Geo.Action(Draw), SYSTEM.ADR(buffer)
            );
          END;
        ELSE
          (* ----------------- ellipse is degenerated into the horizontal line *)
          Geo.SetRect
          (
            r,
            p.ix - VAL(INTEGER, a0), p.ix + VAL(INTEGER, a0),
            p.iy, p.iy
          );
          Draw(r, SYSTEM.ADR(buffer));
        END;
      ELSIF b0 > 0 THEN
        (* ------------------ ellipse is degenerated into the vertical line *)
        Geo.SetRect
        (
          r,
          p.ix, p.ix,
          p.iy - VAL(INTEGER, b0), p.iy + VAL(INTEGER, b0)
        );
        Draw(r, SYSTEM.ADR(buffer));
      ELSE
        (* --------- ellipse is degenerated into the point *)
        Geo.SetRect(r, p.ix, p.ix, p.iy, p.iy);
        Draw(r, SYSTEM.ADR(buffer));
      END;
    END Ellipse;

  (* ===================================== Filled rectangle primitive *)

  PROCEDURE FillRectangle(pm: Scr.PIXMAP; a: Attrs; r: Geo.Rectangle);
    VAR
      buffer: Data;
    BEGIN
      (* ------------------- set the buffer *)
      buffer.attrs := a;
      buffer.pixmap := pm;
      (* -------------------------------- drawing *)
      Draw(r, SYSTEM.ADR(buffer));
    END FillRectangle;

  (* ==================================== Filled circle/ellipse primitive *)

  PROCEDURE FillEllipse(pm: Scr.PIXMAP; a: Attrs; p: Geo.Point; a0, b0: CARDINAL);
    VAR
      buffer: Data;
      r: Geo.Rectangle;
    BEGIN
      (* ------------------- set the buffer *)
      buffer.attrs := a;
      buffer.pixmap := pm;
      (* ------------------- solid/stippleed ellipse *)
      IF a0 > 0 THEN
        (* ------------------- ellipse has non-zero x-halfaxis *)
        IF b0 > 0 THEN
          (* --------- ellipse has internity: check half-axes equality *)
          IF a0 = b0 THEN
            (* --------------------------------------------- it is circle *)
            Geo.TraceFillCircle
            (
              p, a0,
              Geo.Action(Draw), SYSTEM.ADR(buffer)
            );
          ELSE
            (* ---------------------------------------------- it is ellipse *)
            Geo.TraceFillEllipse
            (
              p, a0, b0,
              Geo.Action(Draw), SYSTEM.ADR(buffer)
            );
          END;
        ELSE
          (* ---------------- ellipse is degenerated into the horizontal line *)
          Geo.SetRect
          (
            r,
            p.ix - VAL(INTEGER, a0), p.ix + VAL(INTEGER, a0),
            p.iy, p.iy
          );
          Draw(r, SYSTEM.ADR(buffer));
        END;
      ELSIF b0 > 0 THEN
        (* ------------------- ellipse is degenerated into the vertical line *)
        Geo.SetRect
        (
          r,
          p.ix, p.ix,
          p.iy - VAL(INTEGER, b0),
          p.iy + VAL(INTEGER, b0)
        );
        Draw(r, SYSTEM.ADR(buffer));
      ELSE
        (* --------- ellipse is degenerated into the point *)
        Geo.SetRect(r, p.ix, p.ix, p.iy, p.iy);
        Draw(r, SYSTEM.ADR(buffer));
      END;
    END FillEllipse;

  (* ========================================== Filled polygon primitive *)

  CONST MaxPts = 200;

  PROCEDURE FillPolygon(pm: Scr.PIXMAP; a: Attrs;
            ps: ARRAY OF Geo.Point; first, last: CARDINAL): Res.Status;

    VAR
      buffer: Data;
      r: Geo.Rectangle;
      i, edge, nextEdge, active: CARDINAL;
      y, yMin, yMax, x0, y0, x1, y1, temp: INTEGER;
      x, e: ARRAY [0 .. MaxPts - 1] OF INTEGER;
      xord: ARRAY [0 .. MaxPts - 1] OF CARDINAL;

    PROCEDURE QuickSort(l, r: INTEGER);
      VAR
        i, j, temp, key: INTEGER;
      BEGIN
        WHILE l < r DO
          i := l;
          j := r;
          key := x[xord[j]];
          REPEAT
            WHILE (i < j) AND (x[xord[i]] <= key) DO INC(i) END;
            WHILE (i < j) AND (key <= x[xord[j]]) DO DEC(j) END;
            IF i < j THEN
              temp := xord[i];
              xord[i] := xord[j];
              xord[j] := temp;
            END;
          UNTIL i >= j;
          temp := xord[i];
          xord[i] := xord[r];
          xord[r] := temp;
          IF i - l < r - i THEN
            QuickSort(l, i - 1);
            l := i + 1;
          ELSE
            QuickSort(i + 1, r);
            r := i - 1;
          END;
        END;
      END QuickSort;

    BEGIN
      IF first + 1 < last THEN
        (* ------------------- set the buffer *)
        buffer.attrs := a;
        buffer.pixmap := pm;
        (* --------------------------- correct last index *)
        IF last > MaxPts - 1 THEN last := MaxPts - 1 END;
        (* ---------------------------- find extremal y points *)
        yMin := ps[first].iy;
        yMax := yMin;
        FOR i := first + 1 TO last DO
          IF ps[i].iy < yMin THEN
            yMin := ps[i].iy;
          ELSIF ps[i].iy > yMax THEN
            yMax := ps[i].iy;
          END;
        END;
        (* ------------------------------------------ main cycle *)
        FOR y := yMin TO yMax DO
          (* ------------- initialize counter of active edges *)
          active := 0;
          (* ----------------- determine intersections *)
          FOR edge := first TO last DO
            (* ------------------- set the next node *)
            IF edge # last THEN
              nextEdge := edge + 1;
            ELSE
              nextEdge := 0;
            END;
            (* ------------------ set segment *)
            x0 := ps[edge].ix;
            y0 := ps[edge].iy;
            x1 := ps[nextEdge].ix;
            y1 := ps[nextEdge].iy;
            (* ---------------- swap segment nodes, iff need *)
            IF y0 > y1 THEN
              temp := x0;
              x0 := x1;
              x1 := temp;
              temp := y0;
              y0 := y1;
              y1 := temp;
            END;
            IF y = y0 THEN
              (* -------------------- beginning point *)
              e[edge] := 0;
              x[edge] := x0;
            ELSIF (y0 <= y) AND (y <= y1) THEN
              (* ----------------- determine x-coordinate of intesection *)
              IF x1 >= x0 THEN
                (* -------------------- x increases with y *)
                INC(e[edge], 2 * (x1 - x0));
                WHILE e[edge] > y1 - y0 DO
                  DEC(e[edge], 2 * (y1 - y0));
                  INC(x[edge]);
                END;
              ELSE
                (* ---------------------- x decreases with y *)
                INC(e[edge], 2 * (x0 - x1));
                WHILE e[edge] > y1 - y0 DO
                  DEC(e[edge], 2 * (y1 - y0));
                  DEC(x[edge]);
                END;
              END;
              xord[active] := edge;
              INC(active);
            END;
          END;
          (* ------------ draw the horisontal multi-segment *)
          IF active > 0 THEN
            (* --------------------------- sort *)
            QuickSort(0, active - 1);
            (* ------------------- draw *)
            i := 0;
            r.iyMin := y;
            r.iyMax := y;
            WHILE i < active DO
              r.ixMin := x[xord[i]];
              r.ixMax := x[xord[i + 1]];
              Draw(r, SYSTEM.ADR(buffer));
              INC(i, 2);
            END;
          END;
        END;
      ELSE
        (* ---------------- polygon is degenerated into the line *)
        Geo.TraceLine
        (
          ps[first], ps[last],
          Geo.Action(Draw), SYSTEM.ADR(buffer)
        );
      END;
      RETURN Res.done;
    END FillPolygon;

  PROCEDURE FillPolygonMaxSize(): CARDINAL;
    BEGIN
      RETURN MaxPts;
    END FillPolygonMaxSize;

  (* ======================================= Character/String primitives *)

  PROCEDURE Char(pm: Scr.PIXMAP; a: Attrs; p: Geo.Point; ch: CHAR);
    VAR
      buffer: Data;
      r: Geo.Rectangle;
    BEGIN
      (* ---------------------------- set rectangle *)
      WITH a.font^ DO
        r.ixMin := p.ix;
        r.ixMax := p.ix + VAL(INTEGER, header.charWdt) - 1;
        r.iyMin := p.iy;
        r.iyMax := p.iy + VAL(INTEGER, header.charHgt) - 1;
      END;
      (* ---------------------------------------- set the buffer *)
      buffer.pixmap := pm;
      buffer.attrs := a;
      buffer.attrs.fillOrigin := p;
      buffer.attrs.fillStippleNo := ORD(ch);
      (* ------------------------- character drawing *)
      Draw(r, SYSTEM.ADR(buffer));
    END Char;

  PROCEDURE String(pm: Scr.PIXMAP; a: Attrs; p: Geo.Point; s: ARRAY OF CHAR);
    VAR
      len, j: CARDINAL;
    BEGIN
      (* ------------------------ string length *)
      len := Strings.Length(s);
      j := 0;
      (* ---------------------------------- string drawing *)
      WITH a.font^ DO
        WHILE j < len DO
          Char(pm, a, p, s[j]);
          INC(p.ix, header.charWdt);
          INC(j);
        END;
      END;
    END String;

  (* =========================================== Fonts management *)

  PROCEDURE SafeMult(x, y: CARDINAL; VAR z: CARDINAL): BOOLEAN;
  (* --- safe multiplication of two cardinals *)
    VAR
      long: LONGCARD;
    BEGIN
      long := VAL(LONGCARD, x) * VAL(LONGCARD, y);
      IF long <= VAL(LONGCARD, MAX(CARDINAL)) THEN
        z := VAL(CARDINAL, long);
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END SafeMult;

  PROCEDURE Mess(s: CharSet): CARDINAL;
  (* --- return the number of characters in charset "s" *)
    VAR
      n: CARDINAL;
      ch: CHAR;
    BEGIN
      n := 0;
      FOR ch := MIN(CHAR) TO MAX(CHAR) DO
        IF ch IN s THEN INC(n) END;
      END;
      RETURN n;
    END Mess;

  CONST
    MaxNmbOfStdFonts = 5;

  VAR
    STDFONT: ARRAY [0 .. MaxNmbOfStdFonts - 1] OF FontDescriptor;
    nmbOfStdFonts: CARDINAL;
    topFont: FONT;

  PROCEDURE StdFontsNmb(): CARDINAL;
    BEGIN
      RETURN nmbOfStdFonts;
    END StdFontsNmb;

  PROCEDURE StdFont(n: CARDINAL): FONT;
    BEGIN
      RETURN SYSTEM.ADR(STDFONT[n MOD nmbOfStdFonts]);
    END StdFont;

  PROCEDURE NewEmpty(w, h: CARDINAL; room: CharSet; VAR font: FONT): Res.Status;
    VAR
      newFont: FONT;
      a: SYSTEM.ADDRESS;
      ch: CHAR;
    BEGIN
      IF DrvMem.Alloc(newFont, SIZE(FontDescriptor)) THEN
        WITH newFont^ DO
          (* -------------------- processing of character sizes *)
          header.charWdt := w;
          header.charHgt := h;
          header.byteCharWdt := (header.charWdt - 1) DIV 8 + 1;
          (* ------------------------------ is the font allocatable? *)
          IF
            SafeMult(header.byteCharWdt, h, header.byteCharLen) AND
            SafeMult(header.byteCharLen, Mess(room), header.fontLen) AND
            (header.fontLen > 0)
          THEN
            (* -------------------------------------- yes: test the heap *)
            IF DrvMem.Alloc(header.fontAddr, header.fontLen) THEN
              (* ---------------------------------- set default addresses *)
              SrvLow.FastFill(SYSTEM.ADR(stipple), SIZE(stipple), 0);
              (* ----------------------- load addresses of character stipples *)
              a := header.fontAddr;
              FOR ch := MIN(CHAR) TO MAX(CHAR) DO
                IF ch IN room THEN
                  stipple[ch] := a;
                  a := SrvSys.Add(a, header.byteCharLen);
                END;
              END;
              (* -------------------------------------- fill character stipples *)
              SrvLow.FastFill(header.fontAddr, header.fontLen, 0);
              (* -------------------------------------- set standard flag *)
              standard := FALSE;
              (* --------------- include the font into the font chain *)
              next := topFont;
              topFont := newFont;
              (* ------------------------ normal end *)
              font := newFont;
              RETURN Res.done;
            ELSE
              (* ------------------------ fail due to heap limitations *)
              DrvMem.Free(newFont, SIZE(FontDescriptor));
              RETURN Res.noMemory;
            END;
          ELSE
            (* ----------------------- fail due to bad font size *)
            DrvMem.Free(newFont, SIZE(FontDescriptor));
            RETURN Res.badFontDescription;
          END;
        END;
      ELSE
        RETURN Res.noMemory;
      END;
    END NewEmpty;

  PROCEDURE SetStipple(font: FONT; ch: CHAR; m: SYSTEM.ADDRESS);
    BEGIN
      WITH font^ DO
        IF NOT standard AND (stipple[ch] # NullStipple) THEN
          SrvLow.FastMoveInc(m, stipple[ch], header.byteCharLen);
        END;
      END;
    END SetStipple;

  PROCEDURE Stipple(font: FONT; ch: CHAR; m: SYSTEM.ADDRESS);
    BEGIN
      WITH font^ DO
        IF stipple[ch] # NullStipple THEN
          SrvLow.FastMoveInc(stipple[ch], m, header.byteCharLen);
        END;
      END;
    END Stipple;

  PROCEDURE Chars(font: FONT; VAR room: CharSet);
    VAR
      ch: CHAR;
    BEGIN
      WITH font^ DO
        room := CharSet{};
        FOR ch := MIN(CHAR) TO MAX(CHAR) DO
          IF stipple[ch] # NullStipple THEN INCL(room, ch) END;
        END;
      END;
    END Chars;

  PROCEDURE Dims(font: FONT; VAR w, h: CARDINAL);
    BEGIN
      WITH font^ DO
        w := header.charWdt;
        h := header.charHgt;
      END;
    END Dims;

  PROCEDURE Exist(font: FONT): BOOLEAN;
    BEGIN
      RETURN SrvObj.Exist(SrvObj.OBJECT(topFont), SrvObj.OBJECT(font));
    END Exist;

  PROCEDURE Dispose(VAR font: FONT);
    VAR
      prevFont: FONT;
    BEGIN
      prevFont := FONT(SrvObj.Prev(SrvObj.OBJECT(topFont), SrvObj.OBJECT(font)));
      IF prevFont # NIL THEN
        WITH font^ DO
          IF NOT standard THEN
            (* ------------------------ deallocate the font stipples *)
            DrvMem.Free(header.fontAddr, header.fontLen);
            (* ------------------ exclude the font from the font chain *)
            IF font # topFont THEN
              prevFont^.next := next;
            ELSE
              topFont := next;
            END;
            (* -------------- deallocate the font descriptor *)
            DrvMem.Free(font, SIZE(FontDescriptor));
          END;
        END;
      END;
    END Dispose;

  (* ====================================== Font table procedures *)

  PROCEDURE SetTableFont(index: CARDINAL; font: FONT);
    BEGIN
      fontTable[index MOD FontTableLen] := font;
    END SetTableFont;

  PROCEDURE TableFont(index: CARDINAL): FONT;
    BEGIN
      RETURN fontTable[index MOD FontTableLen];
    END TableFont;

  PROCEDURE TableFontsNmb(): CARDINAL;
    BEGIN
      RETURN FontTableLen;
    END TableFontsNmb;

  (* =========================== Character fonts initialization *)

  PROCEDURE SetFont(dst, src: CARDINAL);
    CONST
      BitsPerByte = 8;
    VAR
      w, h: CARDINAL;
      addr: SYSTEM.ADDRESS;
      ch: CHAR;
    BEGIN
      Scr.FontInfo(src, w, h, addr);
      WITH STDFONT[dst] DO
        WITH header DO
          charWdt := w;
          charHgt := h;
          byteCharWdt := (w - 1) DIV BitsPerByte + 1;
          byteCharLen := byteCharWdt * h;
          FOR ch := MIN(CHAR) TO MAX(CHAR) DO
            stipple[ch] := addr;
            addr := SrvSys.Add(addr, byteCharLen);
          END;
          fontAddr := NIL;
          fontLen := 0;
        END;
        standard := TRUE;
      END;
    END SetFont;

  CONST
    NmbOfHatches = 12;
    StdHatchWdt = 8;
    StdHatchHgt = 8;
    StdByteHatchWdt = (StdHatchWdt - 1) DIV 8 + 1;

  TYPE
    PS = ARRAY [0 .. StdByteHatchWdt * StdHatchHgt - 1] OF SYSTEM.BYTE;
    SS = ARRAY [0 .. NmbOfHatches - 1] OF PS;

  (*%T CMPTS*)

  CONST
    StdHatchHeader = FontHeader
    (
      StdHatchWdt,
      StdHatchHgt,
      StdByteHatchWdt,
      StdByteHatchWdt * StdHatchHgt,
      StdByteHatchWdt * StdHatchHgt * NmbOfHatches,
      NIL
    );
    HatchArray = SS
    (
      (* -------------------------------------- horizontal *)
      PS(0FFH, 0, 0, 0, 0FFH, 0, 0, 0),
      (* -------------------------------------- close horizontal *)
      PS(0FFH, 0, 0FFH, 0, 0FFH, 0, 0FFH, 0),
      (* -------------------------------------- vertical *)
      PS(88H, 88H, 88H, 88H, 88H, 88H, 88H, 88H),
      (* -------------------------------------- close vertical *)
      PS(0AAH, 0AAH, 0AAH, 0AAH, 0AAH, 0AAH, 0AAH, 0AAH),
      (* -------------------------------------- fore diagonal *)
      PS(11H, 22H, 44H, 88H, 11H, 22H, 44H, 88H),
      (* -------------------------------------- light fore diagonal *)
      PS(33H, 66H, 0CCH, 99H, 33H, 66H, 0CCH, 99H),
      (* -------------------------------------- back diagonal *)
      PS(88H, 44H, 22H, 11H, 88H, 44H, 22H, 11H),
      (* -------------------------------------- light back diagonal *)
      PS(0CCH, 66H, 33H, 99H, 0CCH, 66H, 33H, 99H),
      (* -------------------------------------- cross *)
      PS(22H, 22H, 0FFH, 22H, 22H, 22H, 0FFH, 22H),
      (* -------------------------------------- diagonal cross *)
      PS(8H, 14H, 22H, 41H, 80H, 41H, 22H, 14H),
      (* -------------------------------------- dot *)
      PS(88H, 0, 22H, 0, 88H, 0, 22H, 0),
      (* -------------------------------------- close dot *)
      PS(0AAH, 55H, 0AAH, 55H, 0AAH, 55H, 0AAH, 55H)
    );

  (*%E*)

  (*%T CMPSB*)

  CONST
    StdHatchHeader:FontHeader =
    [
      StdHatchWdt,
      StdHatchHgt,
      StdByteHatchWdt,
      StdByteHatchWdt * StdHatchHgt,
      StdByteHatchWdt * StdHatchHgt * NmbOfHatches,
      NIL
    ];
    HatchArray:SS =
    [
      (* -------------------------------------- horizontal *)
      [0FFH, 0, 0, 0, 0FFH, 0, 0, 0],
      (* -------------------------------------- close horizontal *)
      [0FFH, 0, 0FFH, 0, 0FFH, 0, 0FFH, 0],
      (* -------------------------------------- vertical *)
      [88H, 88H, 88H, 88H, 88H, 88H, 88H, 88H],
      (* -------------------------------------- close vertical *)
      [0AAH, 0AAH, 0AAH, 0AAH, 0AAH, 0AAH, 0AAH, 0AAH],
      (* -------------------------------------- fore diagonal *)
      [11H, 22H, 44H, 88H, 11H, 22H, 44H, 88H],
      (* -------------------------------------- light fore diagonal *)
      [33H, 66H, 0CCH, 99H, 33H, 66H, 0CCH, 99H],
      (* -------------------------------------- back diagonal *)
      [88H, 44H, 22H, 11H, 88H, 44H, 22H, 11H],
      (* -------------------------------------- light back diagonal *)
      [0CCH, 66H, 33H, 99H, 0CCH, 66H, 33H, 99H],
      (* -------------------------------------- cross *)
      [22H, 22H, 0FFH, 22H, 22H, 22H, 0FFH, 22H],
      (* -------------------------------------- diagonal cross *)
      [8H, 14H, 22H, 41H, 80H, 41H, 22H, 14H],
      (* -------------------------------------- dot *)
      [88H, 0, 22H, 0, 88H, 0, 22H, 0],
      (* -------------------------------------- close dot *)
      [0AAH, 55H, 0AAH, 55H, 0AAH, 55H, 0AAH, 55H]
    ];

  (*%E*)

  PROCEDURE SetHatchFont(dst: CARDINAL);
    VAR
      a: SYSTEM.ADDRESS;
      j: CARDINAL;
    BEGIN
      WITH STDFONT[dst] DO
        header := StdHatchHeader;
        a := SYSTEM.ADR(HatchArray);
        FOR j := 0 TO NmbOfHatches - 1 DO
          stipple[CHR(j)] := a;
          a := SrvSys.Add(a, header.byteCharLen);
        END;
        SrvLow.FastFill
        (
          SYSTEM.ADR(stipple[CHR(NmbOfHatches)]),
          SIZE(stipple) - NmbOfHatches * SIZE(Entry),
          0
        );
        standard := TRUE;
      END;
    END SetHatchFont;

  (* ===================================== Initialization *)

  VAR j: CARDINAL;

  BEGIN
    (* ---------------------- init array of font descriptors *)
    SrvLow.FastFill(SYSTEM.ADR(STDFONT), SIZE(STDFONT), 0);
    (* -------------------------------- get fonts from screen driver *)
    CASE ScrBas.adapter OF
    | ScrBas.cga:
      nmbOfStdFonts := 2;
      SetFont(0, 0);
    | ScrBas.ega:
      nmbOfStdFonts := 4;
      SetFont(2, 0);
      SetFont(3, 1);
      STDFONT[0] := STDFONT[3];
    | ScrBas.vga:
      nmbOfStdFonts := 5;
      SetFont(2, 0);
      SetFont(3, 1);
      SetFont(4, 2);
      STDFONT[0] := STDFONT[4];
    ELSE
      nmbOfStdFonts := 2;
      SetHatchFont(0);
    END;
    (* -------------------------------- set stipple font *)
    SetHatchFont(1);
    (* ------------------------------------------ set font chain *)
    FOR j := 1 TO nmbOfStdFonts - 1 DO
      STDFONT[j].next := SYSTEM.ADR(STDFONT[j - 1]);
    END;
    topFont := SYSTEM.ADR(STDFONT[nmbOfStdFonts - 1]);
    (* ----------------------------------- set font table *)
    FOR j := 0 TO FontTableLen - 1 DO
      IF j < nmbOfStdFonts THEN
        fontTable[j] := SYSTEM.ADR(STDFONT[j]);
      ELSE
        fontTable[j] := NIL;
      END;
    END;
    (* -------------------- set null point *)
    nullPt.ix := 0;
    nullPt.iy := 0;
  END ScrOut.
