IMPLEMENTATION MODULE ScrPk;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    *************************************************************************
    Hardware:   Super Video Graphics Array
    Compiler:   TopSpeed / Stony Brook Modula-2
    -------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Derived from generic module RD         December 90
    A.Iakovlev  Ver  1.1  New module SrvLow                      December 90
    A.Iakovlev  Ver  1.2  New module SrvLow again                May 91
    A.Iakovlev  Ver  1.3  Use combined module RM                 January 92
    A.Iakovlev  Ver  2.0  Universal module for all SVGA          January 92
    A.Iakovlev  Ver  3.0  Memory bars instead of 64K-segments    August 92
    A.Iakovlev  Ver  4.0  Module Res + procedures New/Dispose    August 92
    A.Iakovlev  Ver  5.0  InitPage/InitPageAddrs/TermPageAddrs   August 92
    A.Iakovlev  Ver  6.0  Renaming                               June 93
    A.Iakovlev  Ver  7.0  Import from SYSTEM                     January 94
    A.Iakovlev  Ver  7.1  No use SHORTCARD type                  February 94
    A.Iakovlev  Ver  7.2  No return structured types             February 94
    A.Iakovlev  Ver  8.0  Color type = Scr.Pixel                 September 94
    A.Iakovlev  Ver  9.0  Bar utilities -> module ScrBar         October 94
    A.Iakovlev  Ver  9.1  Use module Mem for heap control        October 94
    A.Iakovlev  Ver 10.0  High/true-color support                October 94
    A.Iakovlev  Ver 10.1  Address increment for MoveDec          March 95
    A.Iakovlev  Ver 10.2  Use new module ScrBar                  March 95
    A.Iakovlev  Ver 10.3  Use single driver for all Pk8-modes    March 95
    *************************************************************************
    Hardware dependencies:
    ------ byte/pixel resolution
    ------ sequential pixels positions through all the memory banks
    ------ ports to select memory bank
    TopSpeed compiler dependencies:
    ------ options (*$...*)
    ------ predefined type LONGCARD
    ------ single-byte sets
    Stony Brook compiler dependencies:
    ------ predefined type LONGCARD
    ------ single-byte sets
  *)

  IMPORT
    SYSTEM,
    Geo, Res, DrvMem, SrvSys, SrvMat, SrvObj, SrvLow,
    ScrTyp, ScrVar, ScrBar, ScrExt, ScrPk8, ScrPkH, ScrPkT, Scr;

  CONST
    BitsPerByte = 8;
    BytesPerScrBar = 10000H;
    MemBarsPerScrBar = BytesPerScrBar DIV ScrBar.BytesPerBar;

  (* ======================================= Type alias names *)

  TYPE
    BS = ScrTyp.BS;
    WS = ScrTyp.WS;
    BP = ScrTyp.BP;

  (* ====================================================== Globals *)

  VAR
    curMode: Scr.Type;
    clear: PROCEDURE(BP, CARDINAL, CARDINAL, Scr.Pixel);
    pt: PROCEDURE(BP, Scr.Pixel, Scr.Function);
    ptColor: PROCEDURE(BP): Scr.Pixel;
    solidLine: PROCEDURE(BP, CARDINAL, Scr.Pixel, Scr.Function);
    stippledLine: PROCEDURE(BP, CARDINAL, BP, CARDINAL, CARDINAL, Scr.Pixel, Scr.Function);

  (* ======================================================= Utilities *)

  (*$N*)

  PROCEDURE BytesPerPixel(m: Scr.Type): CARDINAL;
    BEGIN
      RETURN (ScrExt.Profiles[m].bitsPerPixel - 1) DIV BitsPerByte + 1;
    END BytesPerPixel;

  PROCEDURE SetOtherFields(pm: PIXMAP; r: Geo.Rectangle);
  (* --- set all the other pixmap fields and include pixmap into the stack *)
    BEGIN
      WITH pm^ DO
        surf := r;
        ix0 := surf.ixMin;
        iy0 := surf.iyMin;
        view.port := surf;
        view.clip := TRUE;
        barLength := ScrBar.BytesPerBar - 1;
        screen := FALSE;
        (* ---------------------------- include into the pixmap stack *)
        next := ScrVar.topPm;
        ScrVar.topPm := pm;
      END;
    END SetOtherFields;

  (*$F*)

  (* =========================================== Mode setting *)

  PROCEDURE SetMode(m: Scr.Type);
    BEGIN
      IF (m = Scr.mode8x320x200) OR ((Scr.mode8x640x400 <= m) AND (m <= Scr.mode8x1280x1024)) THEN
        clear := ScrPk8.Clear;
        pt := ScrPk8.Pt;
        ptColor := ScrPk8.PtColor;
        solidLine := ScrPk8.SolidLine;
        stippledLine := ScrPk8.StippledLine;
      ELSIF (Scr.mode15x320x200 <= m) AND (m <= Scr.mode16x1024x768) THEN
        clear := ScrPkH.Clear;
        pt := ScrPkH.Pt;
        ptColor := ScrPkH.PtColor;
        solidLine := ScrPkH.SolidLine;
        stippledLine := ScrPkH.StippledLine;
      ELSIF (Scr.mode24x320x200 <= m) AND (m <= Scr.mode24x800x600) THEN
        clear := ScrPkT.Clear;
        pt := ScrPkT.Pt;
        ptColor := ScrPkT.PtColor;
        solidLine := ScrPkT.SolidLine;
        stippledLine := ScrPkT.StippledLine;
      END;
      curMode := m;
    END SetMode;

  (* ======================================== Full-pixmap functions *)

  PROCEDURE InitPage(w, h: CARDINAL);
    VAR
      r: Geo.Rectangle;
      bytesPerPage: LONGCARD;
    BEGIN
      WITH ScrVar.SPM DO
        bytesPerPixel := BytesPerPixel(curMode);
        bytesPerPage := VAL(LONGCARD, w) * VAL(LONGCARD, h);
        Geo.SetRect(r, 0, (w DIV bytesPerPixel) - 1, 0, h - 1);
        number := VAL(CARDINAL, bytesPerPage DIV BytesPerScrBar);
        length := VAL(CARDINAL, bytesPerPage - VAL(LONGCARD, number) * BytesPerScrBar);
        barLength := MAX(CARDINAL);
        width := w;
        surf := r;
        view.port := r;
      END;
    END InitPage;

  PROCEDURE InitPageAddrs(sel: CARDINAL);
    BEGIN
      ScrVar.SPM.cont.addr := SrvSys.Addr(sel, 0);
    END InitPageAddrs;

  PROCEDURE TermPageAddrs();
    BEGIN
      ScrVar.SPM.cont.addr := NIL;
    END TermPageAddrs;

  PROCEDURE New(r: Geo.Rectangle; VAR pm: PIXMAP): Res.Status;
    VAR
      newPm: PIXMAP;
      len: LONGCARD;
    BEGIN
      IF DrvMem.Alloc(newPm, SIZE(ScrTyp.PixmapDescriptor)) THEN
        WITH newPm^ DO
          (* --------------------- find width and length parameters *)
          bytesPerPixel := BytesPerPixel(curMode);
          width := VAL(CARDINAL, r.ixMax - r.ixMin + 1) * bytesPerPixel;
          len :=
            VAL(LONGCARD, width) *
            VAL(LONGCARD, r.iyMax - r.iyMin + 1);
          (* --------------------------------- check the number of bars *)
          IF len DIV ScrBar.BytesPerBar < ScrTyp.MaxNmbOfBars THEN
            (* ------------------------------ set parameters of last bar *)
            number := VAL(CARDINAL, len DIV ScrBar.BytesPerBar);
            length := VAL(CARDINAL, len MOD ScrBar.BytesPerBar);
            (* ------------------------ process the case when length = 0 *)
            IF length = 0 THEN
              length := ScrBar.BytesPerBar;
              DEC(number);
            END;
            (* --------------------------------- check the type of pixmap *)
            IF number > 0 THEN
              (* ------------ multi-bar pixmap: attempt to allocate bars *)
              IF ScrBar.Create(newPm) THEN
                (* --------------------------- attempt to allocate buffer *)
                IF DrvMem.Alloc(buffer.addr, width + 1) THEN
                  (* ------------------------------------- success: finish *)
                  SetOtherFields(newPm, r);
                  (* ------------------------------------------ normal exit *)
                  pm := newPm;
                  RETURN Res.done;
                ELSE
                  (* ------------------------- fail: buffer not allocated *)
                  DrvMem.Free(newPm, SIZE(ScrTyp.PixmapDescriptor));
                  RETURN Res.noMemory;
                END;
              ELSE
                (* ------------- fail: directory or bars not allocated *)
                DrvMem.Free(newPm, SIZE(ScrTyp.PixmapDescriptor));
                RETURN Res.noMemory;
              END;
            ELSE
              (* ---------- single-bar pixmap: attempt to allocate content *)
              IF DrvMem.Alloc(cont.addr, length) THEN
                (* ---------------------------- attempt to allocate buffer *)
                IF DrvMem.Alloc(buffer.addr, width + 1) THEN
                  (* ------------------------------------- success: finish *)
                  SetOtherFields(newPm, r);
                  (* ---------------------------------------- normal exit *)
                  pm := newPm;
                  RETURN Res.done;
                ELSE
                  (* ----------------------- fail: buffer not allocated *)
                  DrvMem.Free(cont.addr, length);
                  DrvMem.Free(newPm, SIZE(ScrTyp.PixmapDescriptor));
                  RETURN Res.noMemory;
                END;
              ELSE
                (* --------------------- fail: content not allocated *)
                DrvMem.Free(newPm, SIZE(ScrTyp.PixmapDescriptor));
                RETURN Res.noMemory;
              END;
            END;
          ELSE
            (* ------------------ fail: too large pixmap requested *)
            DrvMem.Free(newPm, SIZE(ScrTyp.PixmapDescriptor));
            RETURN Res.tooLargeRectangle;
          END;
        END;
      ELSE
        (* ------------ fail: pixmap descriptor not allocated *)
        RETURN Res.noMemory;
      END;
    END New;

  PROCEDURE Dispose(VAR pm: PIXMAP);
    VAR
      prevPm: PIXMAP;
    BEGIN
      prevPm := PIXMAP(SrvObj.Prev(SrvObj.OBJECT(ScrVar.topPm), SrvObj.OBJECT(pm)));
      IF prevPm # NIL THEN
        WITH pm^ DO
          IF NOT screen THEN
            (* ----------------- deallocate the buffer *)
            DrvMem.Free(buffer.addr, width + 1);
            (* ----------------------- deallocate the pixmap content *)
            IF number > 0 THEN
              (* --------------------------------------- multi-bar pixmap *)
              ScrBar.Destroy(pm);
            ELSE
              (* ----------------------------------- single-bar pixmap *)
              DrvMem.Free(cont.addr, length);
            END;
            (* ----------------------- exclude from the pixmap stack *)
            IF pm # ScrVar.topPm THEN
              prevPm^.next := next;
            ELSE
              ScrVar.topPm := next;
            END;
            (* -------------- deallocate the pixmap descriptor *)
            DrvMem.Free(pm, SIZE(ScrTyp.PixmapDescriptor));
          END;
        END;
      END;
    END Dispose;

  PROCEDURE Copy(dst, src: PIXMAP);
    VAR
      bar: CARDINAL;
      ptr: BP;
    BEGIN
      WITH dst^ DO
        IF screen THEN
          (* -------------- destination is screen *)
          IF src^.screen THEN
            (* ------------- screen <- screen: impossible case *)
          ELSIF src^.number > 0 THEN
            (* ------------------------------------- screen <- multi *)
            FOR bar := 0 TO src^.number - 1 DO
              IF bar MOD MemBarsPerScrBar = 0 THEN
                ScrExt.SetBank(bar DIV MemBarsPerScrBar);
                ptr := cont;
              END;
              SrvLow.FastMoveInc(src^.cont.dir^[bar], ptr.addr, ScrBar.BytesPerBar);
              INC(ptr.ofs, ScrBar.BytesPerBar);
            END;
            SrvLow.FastMoveInc(src^.cont.dir^[number], ptr.addr, src^.length);
          ELSE
            (* ----------------- screen <- single: impossible case *)
          END;
        ELSIF number > 0 THEN
          (* ----------- destination is multi-bar memory pixmap *)
          IF src^.screen THEN
            (* ---------------------------------------- multi <- screen *)
            FOR bar := 0 TO number - 1 DO
              IF bar MOD MemBarsPerScrBar = 0 THEN
                ScrExt.SetBank(bar DIV MemBarsPerScrBar);
                ptr := src^.cont;
              END;
              SrvLow.FastMoveInc(ptr.addr, cont.dir^[bar], ScrBar.BytesPerBar);
              INC(ptr.ofs, ScrBar.BytesPerBar);
            END;
            SrvLow.FastMoveInc(ptr.addr, cont.dir^[number], length);
          ELSIF src^.number > 0 THEN
            (* ----------------------------------------------- multi <- multi *)
            FOR bar := 0 TO number - 1 DO
              SrvLow.FastMoveInc(src^.cont.dir^[bar], cont.dir^[bar], ScrBar.BytesPerBar);
            END;
            SrvLow.FastMoveInc(src^.cont.dir^[number], cont.dir^[number], length);
          ELSE
            (* ------------------- multi <- single: impossible case *)
          END;
        ELSE
          (* --------- destination is single-bar pixmap *)
          IF src^.screen THEN
            (* -------------- single <- screen: impossible case *)
          ELSIF src^.number > 0 THEN
            (* --------------------- single <- multi: impossible case *)
          ELSE
            (* ----------------------------------------- single <- single *)
            SrvLow.FastMoveInc(src^.cont.addr, cont.addr, length);
          END;
        END;
      END;
    END Copy;

  (* ============================================= Graphics primitives *)

  PROCEDURE Clear(pm: PIXMAP; c: Scr.Pixel);
  (* --- the procedure performs screen clearing for all the screen banks
         ignoring the length of last bank, which can be less than 64K *)
    VAR
      bar: CARDINAL;
    BEGIN
      WITH pm^ DO
        IF screen THEN
          (* ------------------------------- screen pixmap *)
          FOR bar := 0 TO number DO
            ScrExt.SetBank(bar);
            clear(cont, 0, width, c);
          END;
        ELSIF number > 0 THEN
          (* ------------------------------ multi-bar memory pixmap *)
          FOR bar := 0 TO number - 1 DO
            clear(BP(cont.dir^[bar]), ScrBar.BytesPerBar, 0, c);
          END;
          clear(BP(cont.dir^[number]), length, 0, c);
        ELSE
          (* --------------------- single-bar memory pixmap *)
          clear(cont, length, 0, c);
        END;
      END;
    END Clear;

  PROCEDURE Pt(pm: PIXMAP; p: Geo.Point; c: Scr.Pixel; func: Scr.Function);
    VAR
      bp: ScrBar.BarPtr;
    BEGIN
      WITH pm^ DO
        (* --------------------------------------------- clipping *)
        IF view.clip AND NOT Geo.PtInRect(p, view.port) THEN
          RETURN;
        END;
        (* ------------------------------ set the destination pointer *)
        ScrBar.Init(pm, p.ix, p.iy, bp);
        (* ---------------------------------- drawing *)
        pt(bp.ptr, c, func);
      END;
    END Pt;

  PROCEDURE PtColor(pm: PIXMAP; p: Geo.Point): Scr.Pixel;
    VAR
      bp: ScrBar.BarPtr;
    BEGIN
      (* ----------------------- set the destination pointer *)
      ScrBar.Init(pm, p.ix, p.iy, bp);
      (* ------------------- get the point color *)
      RETURN ptColor(bp.ptr);
    END PtColor;

  PROCEDURE SolidRect(pm: PIXMAP; r: Geo.Rectangle; c: Scr.Pixel; func: Scr.Function);
    VAR
      byteCount, lineCount, line, rest: CARDINAL;
      bp: ScrBar.BarPtr;
    BEGIN
      WITH pm^ DO
        (* ---------------------------------------------------- clipping *)
        IF view.clip AND NOT Geo.RectXRectRes(r, view.port, r) THEN
          RETURN;
        END;
        (* ------------------------------------------- set drawing parameters *)
        byteCount := VAL(CARDINAL, r.ixMax - r.ixMin + 1) * bytesPerPixel - 1;
        lineCount := r.iyMax - r.iyMin;
        (* ------------------------ set the destination pointer *)
        ScrBar.Init(pm, r.ixMin, r.iyMin, bp);
        (* ------------------------------- main cycle *)
        line := 0;
        LOOP
          (* ------------------ compute the rest of the bar *)
          rest := barLength - ScrBar.Ofs(bp);
          (* ----------------------------------- check the rest *)
          IF byteCount <= rest THEN
            (* ---------------------- one-bank line: draw the line *)
            solidLine(bp.ptr, byteCount + 1, c, func);
            (* ----------------------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* -------------------------------- increment the pointer *)
            ScrBar.Inc(bp, width);
          ELSE
            (* ------------ multi-bar line: draw the line beginning *)
            solidLine(bp.ptr, rest + 1, c, func);
            (* ---------------------------- increment the pointer *)
            ScrBar.Inc(bp, rest + 1);
            (* ----------------------------- draw the line end *)
            solidLine(bp.ptr, byteCount - rest, c, func);
            (* ------------------ check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ----------------- increment the pointer *)
            ScrBar.Inc(bp, width - rest - 1);
          END;
          (* --------- increment the line counter *)
          INC(line);
        END;
      END;
    END SolidRect;

  PROCEDURE StippledRect(pm: PIXMAP; r: Geo.Rectangle;
            stipple: SYSTEM.ADDRESS; w, h: CARDINAL; p: Geo.Point;
            c: Scr.Pixel; func: Scr.Function);
    VAR
      byteCount, lineCount, line, rest, endLen: CARDINAL;
      stippleOfs0, stippleLineOfs, stippleBitOfs, stippleBitOfsEnd: CARDINAL;
      stipplePtr: BP;
      bp: ScrBar.BarPtr;
    BEGIN
      WITH pm^ DO
        (* --------------------------------------------------- clipping *)
        IF view.clip AND NOT Geo.RectXRectRes(r, view.port, r) THEN
          RETURN;
        END;
        (* ------------------------------------------- set drawing parameters *)
        byteCount := VAL(CARDINAL, r.ixMax - r.ixMin + 1) * bytesPerPixel - 1;
        lineCount := r.iyMax - r.iyMin;
        (* ------------------------------ set stipple parameters *)
        stipplePtr.addr := stipple;
        stippleOfs0 := stipplePtr.ofs;
        stippleLineOfs := SrvMat.Mod(r.iyMin - p.iy, h);
        stippleBitOfs := SrvMat.Mod(r.ixMin - p.ix, w * BitsPerByte);
        (* ----------------------------------- set the destination pointer *)
        ScrBar.Init(pm, r.ixMin, r.iyMin, bp);
        (* -------------------------------- main cycle *)
        line := 0;
        LOOP
          (* -------- set the stipple offset and initial bit *)
          stipplePtr.ofs := stippleLineOfs * w + stippleOfs0;
          (* ------------------------- compute the rest of the bar *)
          rest := barLength - ScrBar.Ofs(bp);
          (* -------------------------------------------- check the rest *)
          IF byteCount <= rest THEN
            (* --------------------------------- one-bank line: draw the line *)
            stippledLine(bp.ptr, byteCount + 1, stipplePtr, stippleBitOfs, w, c, func);
            (* ------------------------------------ check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ---------------------------------- increment the pointer *)
            ScrBar.Inc(bp, width);
          ELSE
            (* ----------------- multi-bar line: draw the line beginning *)
            stippledLine(bp.ptr, rest + 1, stipplePtr, stippleBitOfs, w, c, func);
            (* ------------------------------------------ increment the pointer *)
            ScrBar.Inc(bp, rest + 1);
            (* --------------------- compute the stipple bit offset for the line end *)
            endLen := (rest + 1) DIV bytesPerPixel;
            stippleBitOfsEnd := (stippleBitOfs + endLen) MOD (w * BitsPerByte);
            (* ----------------------------------------------- draw the line end *)
            stippledLine(bp.ptr, byteCount - rest, stipplePtr, stippleBitOfsEnd, w, c, func);
            (* -------------------------------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ---------------------------------- increment the pointer *)
            ScrBar.Inc(bp, width - rest - 1);
          END;
          (* ---------------------- increment the stipple line *)
          stippleLineOfs := (stippleLineOfs + 1) MOD h;
          (* --------------- increment the line counter *)
          INC(line);
        END;
      END;
    END StippledRect;

  PROCEDURE RectCopy(dst: PIXMAP; r: Geo.Rectangle;
            src: PIXMAP; p: Geo.Point; func: Scr.Function);

    VAR
      byteCount, lineCount, line, rest, srcRest: CARDINAL;
      ox, oy, rxBegin, ryBegin, pxBegin, pyBegin: INTEGER;
      forwardX, forwardY, bothScreen, buffNeed: BOOLEAN;
      bp, srcBp: ScrBar.BarPtr;
      buffPtr: BP;

    PROCEDURE LineCopy(n: CARDINAL);
    (* --- intra-bank line copying with length n, counted in bytes *)
      VAR
        storePtr: BP;
      BEGIN
        (* -------------------- onscreen inter-bank copying? *)
        IF buffNeed THEN
          (* --------------------------- yes: set the source bank *)
          ScrExt.SetBank(srcBp.bar);
          (* ------------------------- copy the line into the buffer *)
          IF forwardX THEN
            SrvLow.FastMoveInc(srcBp.ptr.addr, buffPtr.addr, n);
          ELSE
            SrvLow.FastMoveDec(srcBp.ptr.addr, buffPtr.addr, n);
          END;
          (* --------- store the source pointer and set the new pointer *)
          storePtr := srcBp.ptr;
          srcBp.ptr := buffPtr;
          (* ------------------------ set the destination bank *)
          ScrExt.SetBank(bp.bar);
        END;
        (* --------------- copy the line to the destination *)
        IF forwardX THEN
          (* ------------------- in forward direction *)
          IF func = Scr.copy THEN
            (* ------------------------------------ in copy mode *)
            SrvLow.FastMoveInc(srcBp.ptr.addr, bp.ptr.addr, n);
          ELSE
            (* ------------------------------------------- in xor mode *)
            SrvLow.FastMoveXorInc(srcBp.ptr.addr, bp.ptr.addr, n);
          END;
        ELSE
          (* ------------------------- in backward direction *)
          IF func = Scr.copy THEN
            (* ------------------------------------- in copy mode *)
            SrvLow.FastMoveDec(srcBp.ptr.addr, bp.ptr.addr, n);
          ELSE
            (* -------------------------------------------- in xor mode *)
            SrvLow.FastMoveXorDec(srcBp.ptr.addr, bp.ptr.addr, n);
          END;
        END;
        (* ------------ restore (if need) the source pointer *)
        IF buffNeed THEN
          srcBp.ptr := storePtr;
        END;
      END LineCopy;

    PROCEDURE Next(n: CARDINAL);
    (* ------ change both the pointers depending upon X-direction flag *)
      BEGIN
        (* ---------------------------------- change pointers *)
        IF forwardX THEN
          ScrBar.Inc(srcBp, n);
          ScrBar.Inc(bp, n);
        ELSE
          ScrBar.Dec(srcBp, n);
          ScrBar.Dec(bp, n);
        END;
        (* ---------------------- set buffering flag *)
        IF bothScreen THEN
          buffNeed := bp.bar # srcBp.bar;
        END;
      END Next;

    PROCEDURE Prev(n: CARDINAL);
    (* ------ reset both the pointers depending upon X-direction flag *)
      BEGIN
        (* ---------------------------------- change pointers *)
        IF forwardX THEN
          ScrBar.Dec(srcBp, n);
          ScrBar.Dec(bp, n);
        ELSE
          ScrBar.Inc(srcBp, n);
          ScrBar.Inc(bp, n);
        END;
        (* ---------------------- set buffering flag *)
        IF bothScreen THEN
          buffNeed := bp.bar # srcBp.bar;
        END;
      END Prev;

    BEGIN
      WITH dst^ DO
        (* --------------------------- clipping *)
        IF view.clip THEN
          ox := r.ixMin;
          oy := r.iyMin;
          IF Geo.RectXRectRes(r, view.port, r) THEN
            (* ----------------- shift of source base point *)
            INC(p.ix, r.ixMin - ox);
            INC(p.iy, r.iyMin - oy);
          ELSE
            (* ---------- empty intersection *)
            RETURN;
          END;
        END;
        (* -------------------------- set direction flags *)
        forwardX := p.ix >= r.ixMin;
        forwardY := p.iy >= r.iyMin;
        (* ------------------------------------------- set drawing parameters *)
        byteCount := VAL(CARDINAL, r.ixMax - r.ixMin + 1) * bytesPerPixel - 1;
        lineCount := r.iyMax - r.iyMin;
        (* --------------------------- set beginning corners *)
        IF forwardX THEN
          rxBegin := r.ixMin;
          pxBegin := p.ix;
        ELSE
          rxBegin := r.ixMax;
          pxBegin := p.ix + r.ixMax - r.ixMin;
        END;
        IF forwardY THEN
          ryBegin := r.iyMin;
          pyBegin := p.iy;
        ELSE
          ryBegin := r.iyMax;
          pyBegin := p.iy + r.iyMax - r.iyMin;
        END;
        (* ----------------------- set the source pointer *)
        ScrBar.Init(src, pxBegin, pyBegin, srcBp);
        IF NOT forwardX THEN
          INC(srcBp.ptr.ofs, src^.bytesPerPixel - 1);
        END;
        (* ------------------------- set the destination pointer *)
        ScrBar.Init(dst, rxBegin, ryBegin, bp);
        IF NOT forwardX THEN
          INC(bp.ptr.ofs, dst^.bytesPerPixel - 1);
        END;
        (* -------------------------------------- set buffering parameters *)
        bothScreen := screen AND src^.screen;
        buffNeed := bothScreen AND (bp.bar # srcBp.bar);
        IF buffNeed THEN
          buffPtr := dst^.buffer;
          IF NOT forwardX THEN INC(buffPtr.ofs, dst^.width) END;
        END;
        (* ------------------------------------------------- main cycle *)
        line := 0;
        LOOP
          (* ------------------------- compute rests *)
          IF forwardX THEN
            rest := barLength - ScrBar.Ofs(bp);
            srcRest := src^.barLength - ScrBar.Ofs(srcBp);
          ELSE
            rest := ScrBar.Ofs(bp);
            srcRest := ScrBar.Ofs(srcBp);
          END;
          (* ---------------------------------- copy the line *)
          IF byteCount <= rest THEN
            IF byteCount <= srcRest THEN
              (* --------------------------------- intra-bank line *)
              LineCopy(byteCount + 1);
            ELSE
              (* ---------- inter-bank two-part line: src-bank is ended *)
              LineCopy(srcRest + 1);
              Next(srcRest + 1);
              LineCopy(byteCount - srcRest);
              Prev(srcRest + 1);
            END;
          ELSIF byteCount <= srcRest THEN
            (* ------------- inter-bank two-part line: dest-bank is ended *)
            LineCopy(rest + 1);
            Next(rest + 1);
            LineCopy(byteCount - rest);
            Prev(rest + 1);
          ELSIF rest < srcRest THEN
            (* ---------- inter-bank three-part line: both banks are ended *)
            LineCopy(rest + 1);
            Next(rest + 1);
            LineCopy(srcRest - rest);
            Next(srcRest - rest);
            LineCopy(byteCount - srcRest);
            Prev(srcRest + 1);
          ELSIF srcRest < rest THEN
            (* ---------- inter-bank three-part line: both banks are ended *)
            LineCopy(srcRest + 1);
            Next(srcRest + 1);
            LineCopy(rest - srcRest);
            Next(rest - srcRest);
            LineCopy(byteCount - rest);
            Prev(rest + 1);
          ELSE
            (* ---- inter-bank two-part line: banks are equal and ended *)
            LineCopy(rest + 1);
            Next(rest + 1);
            LineCopy(byteCount - rest);
            Prev(rest + 1);
          END;
          (* ------------------------- check the stop condition *)
          IF line = lineCount THEN EXIT END;
          (* -------------- change pointers to the next line *)
          IF forwardY THEN
            ScrBar.Inc(srcBp, src^.width);
            ScrBar.Inc(bp, width);
          ELSE
            ScrBar.Dec(srcBp, src^.width);
            ScrBar.Dec(bp, width);
          END;
          (* ------------------------ set buffering flag *)
          IF bothScreen THEN
            buffNeed := bp.bar # srcBp.bar;
          END;
          (* ----------- increment the line counter *)
          INC(line);
        END;
      END;
    END RectCopy;

  (* =========================== Partitions store/restore functions *)

  PROCEDURE PartsNmb(pm: PIXMAP): CARDINAL;
    BEGIN
      WITH pm^ DO
        IF screen THEN
          RETURN (number + 1) * MemBarsPerScrBar;
        ELSE
          RETURN number + 1;
        END;
      END;
    END PartsNmb;

  PROCEDURE PartSize(pm: PIXMAP; n: CARDINAL): CARDINAL;
    VAR
      psN: CARDINAL;
    BEGIN
      (* ---------- find out the number of parts *)
      psN := PartsNmb(pm);
      (* --------------------- return the size of part *)
      WITH pm^ DO
        IF n MOD psN < psN - 1 THEN
          RETURN ScrBar.BytesPerBar;
        ELSE
          RETURN length;
        END;
      END;
    END PartSize;

  PROCEDURE Store(pm: PIXMAP; n: CARDINAL; s: SYSTEM.ADDRESS);
    VAR
      psN, size: CARDINAL;
      ptr: BP;
    BEGIN
      (* ---------- find out the number of parts and correct "n" *)
      psN := PartsNmb(pm);
      n := n MOD psN;
      (* ------------------------------- store the part *)
      WITH pm^ DO
        IF screen THEN
          (* ------------------- screen pixmap *)
          ptr := cont;
          INC(ptr.ofs, n MOD MemBarsPerScrBar);
          IF n < psN - 1 THEN
            size := ScrBar.BytesPerBar;
          ELSE
            size := length - (length DIV ScrBar.BytesPerBar) * ScrBar.BytesPerBar;
          END;
          ScrExt.SetBank(n DIV MemBarsPerScrBar);
          SrvLow.FastMoveInc(ptr.addr, s, size);
        ELSIF number > 0 THEN
          (* ------------------------------------- multi-bar memory pixmap *)
          IF n < psN - 1 THEN
            size := ScrBar.BytesPerBar;
          ELSE
            size := length;
          END;
          SrvLow.FastMoveInc(cont.dir^[n], s, size);
        ELSE
          (* --------------------------- single-bar memory pixmap *)
          SrvLow.FastMoveInc(cont.addr, s, length);
        END;
      END;
    END Store;

  PROCEDURE Restore(pm: PIXMAP; n: CARDINAL; s: SYSTEM.ADDRESS);
    VAR
      psN, size: CARDINAL;
      ptr: BP;
    BEGIN
      (* ---------- find out the number of parts and correct "n" *)
      psN := PartsNmb(pm);
      n := n MOD psN;
      (* ------------------------------- restore the part *)
      WITH pm^ DO
        IF screen THEN
          (* ------------------- screen pixmap *)
          ptr := cont;
          INC(ptr.ofs, n MOD MemBarsPerScrBar);
          IF n < psN - 1 THEN
            size := ScrBar.BytesPerBar;
          ELSE
            size := length - (length DIV ScrBar.BytesPerBar) * ScrBar.BytesPerBar;
          END;
          ScrExt.SetBank(n DIV MemBarsPerScrBar);
          SrvLow.FastMoveInc(s, ptr.addr, size);
        ELSIF number > 0 THEN
          (* ------------------------------------- multi-bar memory pixmap *)
          IF n < psN - 1 THEN
            size := ScrBar.BytesPerBar;
          ELSE
            size := length;
          END;
          SrvLow.FastMoveInc(s, cont.dir^[n], size);
        ELSE
          (* --------------------------- single-bar memory pixmap *)
          SrvLow.FastMoveInc(s, cont.addr, length);
        END;
      END;
    END Restore;

  END ScrPk.
