IMPLEMENTATION MODULE ScrPk8;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    *************************************************************************
    Hardware:   Super Video Graphics Array
    Compiler:   TopSpeed / Stony Brook Modula-2
    -------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Derived from generic module RD         October 94
    *************************************************************************
    Hardware dependencies:
    ------ byte per pixel resolution
    ------ sequential pixels positions through all the memory banks
    TopSpeed compiler dependencies:
    ------ single-byte sets
    Stony Brook compiler dependencies:
    ------ single-byte sets
  *)

  IMPORT SrvLow, ScrTyp, Scr;

  TYPE BS = ScrTyp.BS;

  CONST BitsPerByte = 8;

  PROCEDURE Clear(ptr: ScrTyp.BP; count, width: CARDINAL; c: Scr.Pixel);
    BEGIN
      IF count = 0 THEN
        SrvLow.SegFill(ptr.seg, BS(c));
      ELSE
        SrvLow.FastFill(ptr.addr, count, BS(c));
      END;
    END Clear;

  PROCEDURE Pt(ptr: ScrTyp.BP; c: Scr.Pixel; rop: Scr.Function);
    BEGIN
      IF rop = Scr.copy THEN
        ptr.addr^ := BS(c);
      ELSE
        ptr.addr^ := ptr.addr^ / BS(c);
      END;
    END Pt;

  PROCEDURE PtColor(ptr: ScrTyp.BP): Scr.Pixel;
    BEGIN
      RETURN VAL(Scr.Pixel, ORD(CHAR(ptr.addr^)));
    END PtColor;

  PROCEDURE SolidLine(ptr: ScrTyp.BP; count: CARDINAL;
            c: Scr.Pixel; rop: Scr.Function);
    BEGIN
      IF rop = Scr.copy THEN
        IF count > 1 THEN
          SrvLow.FastFill(ptr.addr, count, BS(c));
        ELSE
          ptr.addr^ := BS(c);
        END;
      ELSE
        IF count > 1 THEN
          SrvLow.FastFillXor(ptr.addr, count, BS(c));
        ELSE
          ptr.addr^ := ptr.addr^ / BS(c);
        END;
      END;
    END SolidLine;

  PROCEDURE StippledLine(ptr: ScrTyp.BP; count: CARDINAL;
            stipplePtr: ScrTyp.BP; stippleBitOfs, stippleWdt: CARDINAL;
            c: Scr.Pixel; rop: Scr.Function);
    VAR
      pixel, bit, ofs0: CARDINAL;
      stippleLarge: BOOLEAN;
      BM: BS;
    BEGIN
      stippleLarge := (stippleWdt > 1);
      ofs0 := stipplePtr.ofs;
      INC(stipplePtr.ofs, stippleBitOfs DIV BitsPerByte);
      BM := stipplePtr.addr^;
      (* -------------------------------- check the stipple *)
      IF (BM # BS{}) OR stippleLarge THEN
        (* ------------- stipple is nonempty or large: draw it *)
        bit := BitsPerByte - 1 - stippleBitOfs MOD BitsPerByte;
        pixel := 0;
        (* ---------------------------------------- draw the line *)
        LOOP
          (* ------------------------------- change the pixel color *)
          IF bit IN BM THEN
            IF rop = Scr.copy THEN
              ptr.addr^ := BS(c);
            ELSE
              ptr.addr^ := ptr.addr^ / BS(c);
            END;
          END;
          (* -------------------------------- check the stop condition *)
          IF pixel = count - 1 THEN EXIT END;
          (* ---------- change the stipple bit and (if need) read the stipple *)
          IF bit > 0 THEN
            DEC(bit);
          ELSE
            IF stippleLarge THEN
              IF stipplePtr.ofs - ofs0 < stippleWdt - 1 THEN
                INC(stipplePtr.ofs);
              ELSE
                stipplePtr.ofs := ofs0;
              END;
              BM := stipplePtr.addr^;
            END;
            bit := BitsPerByte - 1;
          END;
          (* ---------- increment the destination offset *)
          INC(ptr.ofs);
          (* ------- increment the pixel counter *)
          INC(pixel);
        END;
      END;
    END StippledLine;

  END ScrPk8.
