IMPLEMENTATION MODULE ScrXX;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **********************************************************************
    Compilers:  TopSpeed / Stony Brook Modula-2
    ----------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Derived from all the cursor modules  February 92
    A.Iakovlev  Ver  2.0  Changes in pragmas                   June 92
    A.Iakovlev  Ver  3.0  Use module Res                       August 92
    A.Iakovlev  Ver  4.0  Renaming                             June 93
    A.Iakovlev  Ver  4.1  Import from SYSTEM                   January 94
    A.Iakovlev  Ver  5.0  Use type ScrTxOut.CharSet            January 94
    A.Iakovlev  Ver  5.1  Use module Strings                   January 94
    A.Iakovlev  Ver  6.0  Type DESTINATION -> to DEF-module    February 94
    A.Iakovlev  Ver  6.1  Use module Mem for heap control      October 94
    A.Iakovlev  Ver  6.2  Use new output modules               March 95
    A.Iakovlev  Ver  7.0  Repositioning of parameters          March 95
    A.Iakovlev  Ver  8.0  No line/fill rectangle cursors       March 95
    **********************************************************************
    TopSpeed library dependencies:
    ------ import from Strings: Assign, Insert, Delete, Concat, Length
    Stony Brook library dependencies:
    ------ import from Strings: Assign, Insert, Delete, Concat, Length
  *)

  IMPORT
    SYSTEM, Strings,
    Geo, Res, SrvObj, DrvMem, Scr, ScrOut;

  VAR top: SYSTEM.ADDRESS;

  (* ======================================== Marker cursor *)

  PROCEDURE OutMarker(mc: MARKERCURSOR);
    BEGIN
      WITH mc^ DO
        outProc(dstCurrent, attrCurrent, ptCurrent);
      END;
    END OutMarker;

  PROCEDURE NewMarker(dst: DESTINATION;
            a: ScrOut.Attrs; p: Geo.Point; out: MarkProc;
            VAR mc: MARKERCURSOR): Res.Status;
    VAR
      new: MARKERCURSOR;
    BEGIN
      IF DrvMem.Alloc(new, SIZE(MarkerCursorDescriptor)) THEN
        WITH new^ DO
          (* ---------------------------- set parameters *)
          dstCurrent := dst;
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
          ptCurrent := p;
          outProc := out;
          (* -------------------------------------- set hidden flag *)
          hidden := TRUE;
          (* --------------- include into the marker cursor stack *)
          next := top;
          top := new;
        END;
        mc := new;
        RETURN Res.done;
      ELSE
        RETURN Res.noMemory;
      END;
    END NewMarker;

  PROCEDURE MoveMarker(mc: MARKERCURSOR; p: Geo.Point);
    BEGIN
      WITH mc^ DO
        IF hidden THEN
          ptCurrent := p;
        ELSE
          OutMarker(mc);
          ptCurrent := p;
          OutMarker(mc);
        END;
      END;
    END MoveMarker;

  PROCEDURE ChangeMarker(mc: MARKERCURSOR; a: ScrOut.Attrs);
    BEGIN
      WITH mc^ DO
        IF hidden THEN
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
        ELSE
          OutMarker(mc);
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
          OutMarker(mc);
        END;
      END;
    END ChangeMarker;

  PROCEDURE HideMarker(mc: MARKERCURSOR);
    BEGIN
      WITH mc^ DO
        IF NOT hidden THEN
          OutMarker(mc);
          hidden := TRUE;
        END;
      END;
    END HideMarker;

  PROCEDURE ShowMarker(mc: MARKERCURSOR);
    BEGIN
      WITH mc^ DO
        IF hidden THEN
          OutMarker(mc);
          hidden := FALSE;
        END;
      END;
    END ShowMarker;

  PROCEDURE Marker(mc: MARKERCURSOR; VAR a: ScrOut.Attrs; VAR p: Geo.Point);
    BEGIN
      WITH mc^ DO
        a := attrCurrent;
        p := ptCurrent;
      END;
    END Marker;

  PROCEDURE ExistMarker(mc: MARKERCURSOR): BOOLEAN;
    BEGIN
      RETURN SrvObj.Exist(SrvObj.OBJECT(top), SrvObj.OBJECT(mc));
    END ExistMarker;

  PROCEDURE DisposeMarker(VAR mc: MARKERCURSOR);
    VAR
      prev: MARKERCURSOR;
    BEGIN
      prev := MARKERCURSOR(SrvObj.Prev(SrvObj.OBJECT(top), SrvObj.OBJECT(mc)));
      IF prev # NIL THEN
        WITH mc^ DO
          (* ----------------- erase from the screen *)
          IF NOT hidden THEN
            OutMarker(mc);
          END;
          (* -------------------- exclude from the marker cursor stack *)
          IF mc # top THEN
            prev^.next := next;
          ELSE
            top := next;
          END;
        END;
        (* -------------------- deallocate the marker cursor descriptor *)
        DrvMem.Free(mc, SIZE(MarkerCursorDescriptor));
      END;
    END DisposeMarker;

  (* ========================================= Rectangle Cursor *)

  PROCEDURE OutRectangle(c: RECTCURSOR);
    BEGIN
      WITH c^ DO
        outProc(dstCurrent, attrCurrent, rectCurrent);
      END;
    END OutRectangle;

  PROCEDURE NewRectangle(dst: DESTINATION;
            a: ScrOut.Attrs; r: Geo.Rectangle; out: RectProc;
            VAR c: RECTCURSOR): Res.Status;
    VAR
      new: RECTCURSOR;
    BEGIN
      IF DrvMem.Alloc(new, SIZE(RectDescriptor)) THEN
        WITH new^ DO
          (* ---------------------------- set parameters *)
          dstCurrent := dst;
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
          rectCurrent := r;
          outProc := out;
          (* ------------------------------ set hidden flag *)
          hidden := TRUE;
          (* ---------------- include into the rect stack *)
          next := top;
          top := new;
        END;
        c := new;
        RETURN Res.done;
      ELSE
        RETURN Res.noMemory;
      END;
    END NewRectangle;

  PROCEDURE MoveRectangle(c: RECTCURSOR; r: Geo.Rectangle);
    BEGIN
      WITH c^ DO
        IF hidden THEN
          rectCurrent := r;
        ELSE
          OutRectangle(c);
          rectCurrent := r;
          OutRectangle(c);
        END;
      END;
    END MoveRectangle;

  PROCEDURE ChangeRectangle(c: RECTCURSOR; a: ScrOut.Attrs);
    BEGIN
      WITH c^ DO
        IF hidden THEN
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
        ELSE
          OutRectangle(c);
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
          OutRectangle(c);
        END;
      END;
    END ChangeRectangle;

  PROCEDURE HideRectangle(c: RECTCURSOR);
    BEGIN
      WITH c^ DO
        IF NOT hidden THEN
          OutRectangle(c);
          hidden := TRUE;
        END;
      END;
    END HideRectangle;

  PROCEDURE ShowRectangle(c: RECTCURSOR);
    BEGIN
      WITH c^ DO
        IF hidden THEN
          OutRectangle(c);
          hidden := FALSE;
        END;
      END;
    END ShowRectangle;

  PROCEDURE Rectangle(c: RECTCURSOR; VAR a: ScrOut.Attrs; VAR r: Geo.Rectangle);
    BEGIN
      WITH c^ DO
        a := attrCurrent;
        r := rectCurrent;
      END;
    END Rectangle;

  PROCEDURE ExistRectangle(c: RECTCURSOR): BOOLEAN;
    BEGIN
      RETURN SrvObj.Exist(SrvObj.OBJECT(top), SrvObj.OBJECT(c));
    END ExistRectangle;

  PROCEDURE DisposeRectangle(VAR c: RECTCURSOR);
    VAR
      prev: RECTCURSOR;
    BEGIN
      prev := RECTCURSOR(SrvObj.Prev(SrvObj.OBJECT(top), SrvObj.OBJECT(c)));
      IF prev # NIL THEN
        WITH c^ DO
          (* ---------------- erase from the screen *)
          IF NOT hidden THEN
            OutRectangle(c);
          END;
          (* ----------- exclude froim the rectangle stack *)
          IF c # top THEN
            prev^.next := next;
          ELSE
            top := next;
          END;
        END;
        (* ------------- deallocate the rectangle descriptor *)
        DrvMem.Free(c, SIZE(RectDescriptor));
      END;
    END DisposeRectangle;

  (* ========================================= Character cursor *)

  PROCEDURE OutChar(c: CHARCURSOR);
    BEGIN
      WITH c^ DO
        outProc(dstCurrent, attrCurrent, ptCurrent, charCurrent);
      END;
    END OutChar;

  PROCEDURE NewChar(dst: DESTINATION;
            a: ScrOut.Attrs; p: Geo.Point; ch: CHAR; out: CharProc;
            VAR c: CHARCURSOR): Res.Status;
    VAR
      new: CHARCURSOR;
    BEGIN
      IF DrvMem.Alloc(new, SIZE(CharCursorDescriptor)) THEN
        WITH new^ DO
          (* ---------------------------- set parameters *)
          dstCurrent := dst;
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
          ptCurrent := p;
          charCurrent := ch;
          outProc := out;
          (* ------------------------------------- set hidden flag *)
          hidden := TRUE;
          (* ------------ include into the character cursor chain *)
          next := top;
          top := new;
          (* ---------------------- normal end *)
          c := new;
          RETURN Res.done;
        END;
      ELSE
        (* --------------- fail due to memory limitations *)
        RETURN Res.noMemory;
      END;
    END NewChar;

  PROCEDURE MoveChar(c: CHARCURSOR; p: Geo.Point);
    BEGIN
      WITH c^ DO
        IF hidden THEN
          ptCurrent := p;
        ELSE
          OutChar(c);
          ptCurrent := p;
          OutChar(c);
        END;
      END;
    END MoveChar;

  PROCEDURE ChangeChar(c: CHARCURSOR; a: ScrOut.Attrs);
    BEGIN
      WITH c^ DO
        IF hidden THEN
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
        ELSE
          OutChar(c);
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
          OutChar(c);
        END;
      END;
    END ChangeChar;

  PROCEDURE ReplaceChar(c: CHARCURSOR; ch: CHAR);
    BEGIN
      WITH c^ DO
        IF hidden THEN
          charCurrent := ch;
        ELSE
          OutChar(c);
          charCurrent := ch;
          OutChar(c);
        END;
      END;
    END ReplaceChar;

  PROCEDURE HideChar(c: CHARCURSOR);
    BEGIN
      WITH c^ DO
        IF NOT hidden THEN
          OutChar(c);
          hidden := TRUE;
        END;
      END;
    END HideChar;

  PROCEDURE ShowChar(c: CHARCURSOR);
    BEGIN
      WITH c^ DO
        IF hidden THEN
          OutChar(c);
          hidden := FALSE;
        END;
      END;
    END ShowChar;

  PROCEDURE Char(c: CHARCURSOR; VAR a: ScrOut.Attrs; VAR p: Geo.Point; VAR ch: CHAR);
    BEGIN
      WITH c^ DO
        a := attrCurrent;
        p := ptCurrent;
        ch := charCurrent;
      END;
    END Char;

  PROCEDURE ExistChar(c: CHARCURSOR): BOOLEAN;
    BEGIN
      RETURN SrvObj.Exist(SrvObj.OBJECT(top), SrvObj.OBJECT(c));
    END ExistChar;

  PROCEDURE DisposeChar(VAR c: CHARCURSOR);
    VAR
      prev: CHARCURSOR;
    BEGIN
      prev := CHARCURSOR(SrvObj.Prev(SrvObj.OBJECT(top), SrvObj.OBJECT(c)));
      IF prev # NIL THEN
        WITH c^ DO
          (* ----------------- erase from the screen *)
          IF NOT hidden THEN
            OutChar(c);
          END;
          (* -------------------- exclude from the character cursor chain *)
          IF c # top THEN
            prev^.next := next;
          ELSE
            top := next;
          END;
        END;
        (* ------------ deallocate the character cursor descriptor *)
        DrvMem.Free(c, SIZE(CharCursorDescriptor));
      END;
    END DisposeChar;

  (* =========================================== String cursor *)

  PROCEDURE OutString(c: STRINGCURSOR);
    VAR
      p: Geo.Point;
      len, j: CARDINAL;
    BEGIN
      WITH c^ DO
        len := Strings.Length(strCurrent);
        IF len > 0 THEN
          p := ptCurrent;
          FOR j := 0 TO len - 1 DO
            outProc(dstCurrent, attrCurrent, p, strCurrent[j]);
            INC(p.ix, charWdt);
          END;
        END;
      END;
    END OutString;

  PROCEDURE NewString(dst: DESTINATION;
            a: ScrOut.Attrs; p: Geo.Point; s: ARRAY OF CHAR; out: CharProc;
            VAR c: STRINGCURSOR): Res.Status;
    VAR
      new: STRINGCURSOR;
      j, strLen: CARDINAL;
    BEGIN
      IF DrvMem.Alloc(new, SIZE(StringCursorDescriptor)) THEN
        WITH new^ DO
          (* ----------------------- set parameters *)
          dstCurrent := dst;
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
          ptCurrent := p;
          Strings.Assign(s, strCurrent);
          outProc := out;
          (* -------------------------------------- get character sizes *)
          ScrOut.Dims(attrCurrent.font, charWdt, charHgt);
          (* -------------------------------------- set hidden flag *)
          hidden := TRUE;
          (* ------------- include into the string cursor chain *)
          next := top;
          top := new;
          (* --------------------------- normal end *)
          c := new;
          RETURN Res.done;
        END;
      ELSE
        (* --------------- fail due to memory limitations *)
        RETURN Res.noMemory;
      END;
    END NewString;

  PROCEDURE MoveString(c: STRINGCURSOR; p: Geo.Point);
    BEGIN
      WITH c^ DO
        IF hidden THEN
          ptCurrent := p;
        ELSE
          OutString(c);
          ptCurrent := p;
          OutString(c);
        END;
      END;
    END MoveString;

  PROCEDURE ChangeString(c: STRINGCURSOR; a: ScrOut.Attrs);
    BEGIN
      WITH c^ DO
        IF hidden THEN
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
        ELSE
          OutString(c);
          attrCurrent := a;
          attrCurrent.function := Scr.xor;
          OutString(c);
        END;
      END;
    END ChangeString;

  PROCEDURE LengthString(c: STRINGCURSOR): CARDINAL;
    BEGIN
      RETURN Strings.Length(c^.strCurrent);
    END LengthString;

  PROCEDURE ModSubString(c: STRINGCURSOR; s: ARRAY OF CHAR; pos: CARDINAL; ins: BOOLEAN);
    VAR
      p: Geo.Point;
      strLen, len, pos1, j, jmax: CARDINAL;
      ch: CHAR;
    BEGIN
      (* -------------------- find length of substring *)
      len := Strings.Length(s);
      (* ------------------------------ check this length *)
      IF len > 0 THEN
        (* ----------------- correct length to avoid overflow *)
        IF len > MaxStrLen THEN len := MaxStrLen END;
        (* --------------------------------------- correct string *)
        WITH c^ DO
          (* ---------------------- find the current length of string *)
          strLen := Strings.Length(strCurrent);
          (* --------------------- check position relatively this length *)
          IF pos < strLen THEN
            (* ----------- position is within the string: check insert flag *)
            IF ins THEN
              (* ------------------------------------ in insert mode *)
              IF NOT hidden THEN
                (* ---------------------------- set parameters *)
                jmax := strLen + len - 1;
                IF jmax >= MaxStrLen THEN jmax := MaxStrLen - 1 END;
                pos1 := pos + len;
                (* ----------------------------------- set base point *)
                p.ix := ptCurrent.ix + INTEGER(jmax * charWdt);
                p.iy := ptCurrent.iy;
                (* -------------------------------------- insert on pixmap *)
                FOR j := jmax TO pos BY - 1 DO
                  IF j < strLen THEN
                    outProc(dstCurrent, attrCurrent, p, strCurrent[j]);
                  END;
                  IF j >= pos1 THEN ch := strCurrent[j - len] ELSE ch := s[j - pos] END;
                  outProc(dstCurrent, attrCurrent, p, ch);
                  DEC(p.ix, charWdt);
                END;
              END;
              (* ------------------- correct string in the memory *)
              Strings.Insert(s, strCurrent, pos);
            ELSE
              (* ------------------------------------ in replace mode *)
              IF NOT hidden THEN
                (* ----------------------------- set parameters *)
                jmax := pos + len - 1;
                IF jmax >= MaxStrLen THEN jmax := MaxStrLen - 1 END;
                (* ---------------------------------------- set base point *)
                p.ix := ptCurrent.ix + INTEGER(pos * charWdt);
                p.iy := ptCurrent.iy;
                (* ---------------------------------------- replace on pixmap *)
                FOR j := pos TO jmax DO
                  IF j < strLen THEN
                    outProc(dstCurrent, attrCurrent, p, strCurrent[j]);
                  END;
                  outProc(dstCurrent, attrCurrent, p, s[j - pos]);
                  INC(p.ix, charWdt);
                END;
              END;
              (* ------------------ correct string in the memory *)
              Strings.Delete(strCurrent, pos, len);
              Strings.Insert(s, strCurrent, pos);
            END;
          ELSE
            (* ------- position is after the end of string: redraw the string *)
            IF NOT hidden THEN
              (* ---------------------- set parameters *)
              jmax := strLen + len - 1;
              IF jmax >= MaxStrLen THEN jmax := MaxStrLen - 1 END;
              (* -------------------------------------------- set base point *)
              p.ix := ptCurrent.ix + INTEGER(strLen * charWdt);
              p.iy := ptCurrent.iy;
              (* --------------------------------------- append on pixmap *)
              FOR j := strLen TO jmax DO
                outProc(dstCurrent, attrCurrent, p, s[j - strLen]);
                INC(p.ix, charWdt);
              END;
            END;
            (* --------------- correct string in the memory *)
            Strings.Concat(strCurrent, s, strCurrent);
          END;
        END;
      END;
    END ModSubString;

  PROCEDURE DelSubString(c: STRINGCURSOR; pos, len: CARDINAL);
    VAR
      p: Geo.Point;
      strLen, j: CARDINAL;
    BEGIN
      IF len > 0 THEN
        (* ----------------- correct length to avoid overflow *)
        IF len > MaxStrLen THEN len := MaxStrLen END;
        (* --------------------------------------- correct string *)
        WITH c^ DO
          (* ---------------------- find the current length of string *)
          strLen := Strings.Length(strCurrent);
          (* --------------------- check position relatively this length *)
          IF pos < strLen THEN
            (* ---------- position is within the string: redraw the string *)
            IF NOT hidden THEN
              (* ------------------------------------------- set base point *)
              p.ix := ptCurrent.ix + INTEGER(pos * charWdt);
              p.iy := ptCurrent.iy;
              (* ------------------------------------------ delete on pixmap *)
              FOR j := pos TO strLen - 1 DO
                outProc(dstCurrent, attrCurrent, p, strCurrent[j]);
                IF j + len < strLen THEN
                  outProc(dstCurrent, attrCurrent, p, strCurrent[j + len]);
                END;
                INC(p.ix, charWdt);
              END;
            END;
            (* ------------------------ correct string in memory *)
            Strings.Delete(strCurrent, pos, len);
          END;
        END;
      END;
    END DelSubString;

  PROCEDURE SkipString(c: STRINGCURSOR; pos: CARDINAL; set: ScrOut.CharSet; forward: BOOLEAN): CARDINAL;
    VAR
      strLen: CARDINAL;
    BEGIN
      WITH c^ DO
        strLen := Strings.Length(strCurrent);
        IF strLen > 0 THEN
          IF forward THEN
            WHILE (pos < strLen) AND (strCurrent[pos] IN set) DO INC(pos) END;
          ELSE
            IF pos >= strLen THEN pos := strLen - 1 END;
            WHILE (pos > 0) AND (strCurrent[pos] IN set) DO DEC(pos) END;
          END;
        END;
      END;
      RETURN pos;
    END SkipString;

  PROCEDURE HideString(c: STRINGCURSOR);
    BEGIN
      WITH c^ DO
        IF NOT hidden THEN
          OutString(c);
          hidden := TRUE;
        END;
      END;
    END HideString;

  PROCEDURE ShowString(c: STRINGCURSOR);
    BEGIN
      WITH c^ DO
        IF hidden THEN
          OutString(c);
          hidden := FALSE;
        END;
      END;
    END ShowString;

  PROCEDURE String(c: STRINGCURSOR; VAR a: ScrOut.Attrs; VAR p: Geo.Point; VAR s: ARRAY OF CHAR);
    BEGIN
      WITH c^ DO
        a := attrCurrent;
        p := ptCurrent;
        Strings.Assign(strCurrent, s);
      END;
    END String;

  PROCEDURE ExistString(c: STRINGCURSOR): BOOLEAN;
    BEGIN
      RETURN SrvObj.Exist(SrvObj.OBJECT(top), SrvObj.OBJECT(c));
    END ExistString;

  PROCEDURE DisposeString(VAR c: STRINGCURSOR);
    VAR
      prev: STRINGCURSOR;
    BEGIN
      prev := STRINGCURSOR(SrvObj.Prev(SrvObj.OBJECT(top), SrvObj.OBJECT(c)));
      IF prev # NIL THEN
        WITH c^ DO
          (* ----------------------- erase from the screen *)
          IF NOT hidden THEN
            OutString(c);
          END;
          (* -------------------- exclude from the string cursor stack *)
          IF c # top THEN
            prev^.next := next;
          ELSE
            top := next;
          END;
        END;
        (* ------------ deallocate the string cursor descriptor *)
        DrvMem.Free(c, SIZE(StringCursorDescriptor));
      END;
    END DisposeString;

  BEGIN
    top := NIL;
  END ScrXX.
